# This example is for a dashboard user to modify data in a PostgreSQL DB table using 
# a modern grid configured with transaction functions. The workflow will rely on this 
# script to make actual modifications in the table. The table named USERS has 4 
# columns USER_ID(Int), USER_LOGIN(Varchar), HIRED_DATE(Date) and EMAIL(VarChar) 
# with USER_ID as the primary. Data changes for HIRED_DATE (Date) and EMAIL(VarChar) 
# are optional.

# In this template we use the library 'psycopg2' (installed in runtime as
# 'psycopg2-binary') to connect with the PostgreSQL database, submit atomic requests 
# and execute operations. For using other types of database, make sure to use
# their methods correctly. 

# Python libraries to handle SQL connections are not installed on the environment by
# default so it is necessary to choose runtime which have such library installed in the
# settings of the script. Additionally it is necessary to specify correct `Network
# Access` to allow script to communicate with some other IP then IP of the environment.
import psycopg2
import datetime


# helper method to handle not adding None values to the query
def append_not_none_value(array, value):
    if value is not None:
        array.append(value)
        return True
    return False

# Helper method to return datetime object in correct format for SQL operations
def parseDateTimeValue(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000, 
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%m/%d/%Y",
        "%m/%d/%Y %H:%M:%S",
        "%Y-%d-%m %H:%M:%S",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m-%d-%Y"
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.datetime.strptime(dateValue, date_format)
            break
        except:
            pass

    if (parsed_datetime == None):
        raise Exception("Wrong Type of the datatime input value convertion!!!!")
    else:
        return parsed_datetime


try:
    # create connection to database. If you want to use other way then providing
    # user and password please check documentation of this method 
    connection = None
    connection = psycopg2.connect(
        host=$host,
        port=$port,
        database=$database,
        user=$user_name,
        password=$user_password
    )

    # Prepare query for database operations. Due to the fact that for some rows
    # we might want to pass `None` value we need some additional logic. This logic
    # require to create query during the time of reading passed properties.

    # Parameterized query for edit the existing row
    # We assume that when updating we must pass only `USER_LOGIN` and the other
    # properties are optional. Update is done based on `USER_ID` and `WHERE` clause
    # is added in the end of preparing query for each row
    tmp_sql_update_query = """
        UPDATE USERS SET USER_LOGIN = %s
    """

    # create cursor to execute operations on database
    cursor = connection.cursor()
    
    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`.
    for index in range(len($user_login)):
        optional_values_array = []
        optional_values_in_update_query = ''
        # when array with values for each property is not `None` and the value
        # for particular row is not `None` then we add it 
        if $hired_date and $hired_date[index]:
            optional_values_array.append(parseDateTimeValue($hired_date[index]))
            optional_values_in_update_query += f",HIRED_DATE = %s"
        if $email and append_not_none_value(optional_values_array, $email[index]):
            optional_values_in_update_query += f",EMAIL = %s"

        # final UPDATE query is created based on initial UPDATE query and the part
        # created for particular row based on properties passed for this row
        # after adding optional properties we need to add `WHERE` clause
        # which in this case is based on `USER_ID`
        optional_values_array.append(int($user_id[index]))
        optional_values_in_update_query += f" WHERE USER_ID = %s"

        sql_update_tuple = (
            $user_login[index],
        ) + tuple(optional_values_array)
        sql_update_query = tmp_sql_update_query + optional_values_in_update_query
        # execute update query
        cursor.execute(sql_update_query, sql_update_tuple)

    # commit all changes when there weren't any errors
    connection.commit()

# error handling
except psycopg2.Error as e:
    # in case of any error we have to rollback all of the changes which
    # has been made on database
    if connection:
        connection.rollback()
    print("Error:", e)

finally:
    # close the database connection
    if connection:
        connection.close()
        


from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting bookmarks from shortcut {shortcut_id}")
def get_bookmarks_from_shortcut(connection, shortcut_id, error_msg=None):
    """Get a Bookmark list from a Shortcut Object.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id (string): Shortcut  ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint=f'/api/shortcuts/{shortcut_id}/bookmarks')


@ErrorHandler(err_msg="Error updating information for shortcut {shortcut_id}")
def update_information_for_shortcut(connection, shortcut_id, body, error_msg=None):
    """Update info for a shortcut.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id (string): The ID of the document shortcut to execute
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(endpoint=f'/api/shortcuts/{shortcut_id}', json=body)


@ErrorHandler(err_msg="Error adding a new bookmark.")
def refresh_document_instance(connection, error_msg=None):
    """Add a new bookmark into current shortcut object.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(endpoint='/api/bookmarks')


@ErrorHandler(err_msg="Error deleting bookmarks.")
def delete_bookmarks(connection, shortcut_id: str, bookmark_ids: list, error_msg=None):
    """Bulk deletion of bookmarks.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id: Shortcut ID
        bookmark_ids: IDs of bookmarks to be deleted
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {'shortcutId': shortcut_id, 'bookmarkIds': bookmark_ids}
    return connection.delete(endpoint='/api/bookmarks', json=body)


@ErrorHandler(err_msg="Error deleting bookmark with ID {bookmark_id}")
def delete_single_bookmark(
    connection, shortcut_id: str, bookmark_id: str, error_msg=None
):
    """Delete a bookmark.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id: Shortcut ID
        bookmark_id (string): Bookmark ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {'shortcutId': shortcut_id}
    endpoint = f'/api/bookmarks/{bookmark_id}'
    return connection.delete(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error updating bookmark {bookmark_id}")
def update_bookmark(connection, bookmark_id, body, error_msg=None):
    """Update a bookmark.

    Args:
        connection: MicroStrategy REST API connection object
        bookmark_id (string): Bookmark ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/bookmarks/{bookmark_id}'
    return connection.put(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error adding bookmark {bookmark_name}")
def add_bookmark(
    connection, bookmark_name, instance_id, shortcut_id, error_msg=None
):  # shortcut needs to be created to add a bookmark!
    """Update a bookmark.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {'name': bookmark_name, 'instanceId': instance_id}
    endpoint = '/api/bookmarks'
    return connection.post(endpoint=endpoint, json=body)


@ErrorHandler(err_msg="Error getting shortcuts for document {document_id}")
def get_document_shortcut(connection, document_id, instance_id, error_msg=None):
    """Retrieve a published shortcut from the document definition.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document  ID
        instance_id (string): Instance  ID
        error_msg (string, optional): Custom Error Message for Error Handling


    Returns:
        Complete HTTP response object.
    """
    endpoint = f'/api/documents/{document_id}/instances/{instance_id}/shortcut'
    return connection.get(endpoint=endpoint)



# NOSONAR
import logging
from collections import defaultdict
from collections.abc import Iterable
from enum import auto
from typing import TYPE_CHECKING, Optional

from mstrio import config
from mstrio.api import contacts
from mstrio.users_and_groups.contact_group import ContactGroup
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import DeleteMixin, EntityBase, auto_match_args_entity
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import (
    Dictable,
    camel_to_snake,
    delete_none_values,
    fetch_objects,
    get_args_from_func,
    get_default_args_from_func,
    get_objects_id,
)
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.distribution_services.device.device import Device

logger = logging.getLogger(__name__)


class ContactDeliveryType(AutoName):
    EMAIL = auto()
    FILE = auto()
    PRINTER = auto()
    FTP = auto()
    MOBILE_ANDROID = auto()
    MOBILE_IPHONE = auto()
    MOBILE_IPAD = auto()
    UNSUPPORTED = auto()


class ContactAddress(Dictable):
    """Representation of contact address object

    Attributes:
        id: id of contact address, optional
        name: contact address' name
        physical_address: physical address of contact
        delivery_type: object of type ContactDeliveryType
        is_default: specifies if address is default, optional,
            default value: False
        device: instance of Device or string (containing device's id),
            if device is a string, connection is required
        connection: instance of Connection, optional,
            is required if device is string
    """

    @staticmethod
    def __device_from_dict(source, connection):
        from mstrio.distribution_services.device.device import Device

        return Device.from_dict(source, connection)

    _FROM_DICT_MAP = {
        'delivery_type': ContactDeliveryType,
        'device': __device_from_dict,
    }

    def __init__(
        self,
        name: str,
        physical_address: str,
        delivery_type: ContactDeliveryType | str,
        device: 'Device | str',
        id: str | None = None,
        is_default: bool = False,
        connection: Optional['Connection'] = None,
    ):
        self.id = id
        self.name = name
        self.physical_address = physical_address
        self.is_default = is_default

        self.delivery_type = (
            delivery_type
            if isinstance(delivery_type, ContactDeliveryType)
            else ContactDeliveryType(delivery_type)
        )

        from mstrio.distribution_services.device.device import Device

        if isinstance(device, Device):
            self.device = device
        else:
            if not connection:
                raise ValueError(
                    'Argument: connection is required if device is a string'
                )

            self.device = Device(connection, id=device)

    def __repr__(self) -> str:
        param_dict = auto_match_args_entity(
            self.__init__, self, exclude=['self'], include_defaults=False
        )

        params = [
            (
                f"{param}={self.delivery_type}"
                if param == 'delivery_type'
                else f'{param}={repr(value)}'
            )
            for param, value in param_dict.items()
        ]
        formatted_params = ', '.join(params)

        return f'ContactAddress({formatted_params})'

    def to_dict(self, camel_case=True) -> dict:
        result = {
            'name': self.name,
            'id': self.id,
            'physicalAddress': self.physical_address,
            'deliveryType': self.delivery_type.value,
            'deviceId': self.device.id,
            'deviceName': self.device.name,
            'isDefault': self.is_default,
        }

        return result if camel_case else camel_to_snake(result)

    @classmethod
    def from_dict(cls, source, connection, to_snake_case=True) -> 'ContactAddress':
        source = source.copy()

        device_id = source.pop('deviceId')
        device_name = source.pop('deviceName')

        source['device'] = {'id': device_id, 'name': device_name}

        return super().from_dict(source, connection, to_snake_case)


@method_version_handler('11.3.0100')
def list_contacts(
    connection: 'Connection',
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list['Contact'] | list[dict]:
    """Get all contacts as list of Contact objects or dictionaries.

    Optionally filter the contacts by specifying filters.

    Args:
        connection: MicroStrategy connection object
        to_dictionary: If True returns a list of contact dicts,
            otherwise returns a list of contact objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters:
            ['id', 'name', 'description', 'enabled', 'linked_user']
    """

    return Contact._list_contacts(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@class_version_handler('11.3.0100')
class Contact(EntityBase, DeleteMixin):
    """Object representation of Microstrategy Contact object

    Attributes:
        name: contact's name
        id: contact's id
        description: contact's description
        enabled: specifies if a contact is enabled
        linked_user: user linked to contact, instance of User
        contact_addresses: list of contact's addresses,
            instances of ContactAddress
        memberships: list of Contact Groups that the contact belongs to
        connection: instance of Connection class, represents connection
                    to MicroStrategy Intelligence Server
    """

    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'linked_user': User.from_dict,
        'contact_addresses': [ContactAddress.from_dict],
        'memberships': [ContactGroup.from_dict],
    }
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'enabled',
            'linked_user',
            'memberships',
            'contact_addresses',
        ): contacts.get_contact
    }
    _API_DELETE = staticmethod(contacts.delete_contact)
    _API_PATCH = {
        (
            'name',
            'description',
            'enabled',
            'linked_user',
            'contact_addresses',
            'memberships',
        ): (contacts.update_contact, 'put')
    }
    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'enabled': bool,
        'linked_user': dict,
        'contact_addresses': list,
        'memberships': list,
    }

    def __init__(
        self,
        connection: 'Connection',
        id: str | None = None,
        name: str | None = None,
    ):
        """Initialize Contact object by passing id or name.
        When `id` is provided, name is omitted.

        Args:
            connection: MicroStrategy connection object
            id: ID of Contact
            name: name of Contact
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            result = Contact._list_contacts(
                connection=connection, name=name, to_dictionary=True
            )

            if result:
                object_data = result[0]
                object_data['connection'] = connection
                self._init_variables(**object_data)
            else:
                raise ValueError(f"There is no Contact named: '{name}'")
        else:
            super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.description = kwargs.get('description')
        self.enabled = kwargs.get('enabled')

        linked_user = kwargs.get("linked_user")
        self.linked_user = (
            User.from_dict(linked_user, self.connection) if linked_user else None
        )

        addresses = kwargs.get('contact_addresses')
        self.contact_addresses = (
            [
                ContactAddress.from_dict(address, self.connection)
                for address in addresses
            ]
            if addresses
            else None
        )

        memberships = kwargs.get('memberships')
        self.memberships = (
            [ContactGroup.from_dict(m, self.connection) for m in memberships]
            if memberships
            else None
        )

    @classmethod
    @method_version_handler('11.3.0200')
    def create(
        cls,
        connection: 'Connection',
        name: str,
        linked_user: 'User | str',
        contact_addresses: Iterable['ContactAddress | dict'],
        description: str | None = None,
        enabled: bool = True,
    ) -> 'Contact':
        """Create a new contact.

        Args:
            connection: MicroStrategy connection object
                returned by `connection.Connection()`
            name: contact name
            linked_user: user linked to contact
            contact_addresses: list of contact addresses
            description: description of contact
            enabled: specifies if contact should be enabled
        Returns:
            Contact object
        """
        body = {
            'name': name,
            'description': description,
            'enabled': enabled,
            'linkedUser': {'id': get_objects_id(linked_user, User)},
            'contactAddresses': [
                address.to_dict() if isinstance(address, ContactAddress) else address
                for address in contact_addresses
            ],
        }
        body = delete_none_values(body, recursion=True)
        response = contacts.create_contact(connection, body).json()

        if config.verbose:
            logger.info(
                f"Successfully created contact named: '{name}' with ID: "
                f"'{response['id']}'"
            )

        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        enabled: bool | None = None,
        linked_user: Optional['User | str'] = None,
        contact_addresses: Iterable['ContactAddress | dict'] | None = None,
    ):
        """Update properties of a contact

        Args:
           name: name of a contact
           description: description of a contact
           enabled: specifies if a contact is enabled
           linked_user: an object of class User linked to the contact
           contact_addresses: list of contact addresses
        """
        linked_user = {'id': get_objects_id(linked_user, User)} if linked_user else None

        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        defaults_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = defaultdict(dict)

        for property_key in defaults_dict:
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        self._alter_properties(**properties)

    @classmethod
    def _list_contacts(
        cls,
        connection: 'Connection',
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list['Contact'] | list[dict]:
        """Get all contacts as list of Contact objects or dictionaries.

        Optionally filter the contacts by specifying filters.

        Args:
            connection: MicroStrategy connection object
            to_dictionary: If True returns a list of contact dicts,
               otherwise returns a list of contact objects
           limit: limit the number of elements returned. If `None` (default),
               all objects are returned.
           **filters: Available filter parameters:
               ['id', 'name', 'description', 'enabled']
        """

        objects = fetch_objects(
            connection=connection,
            api=contacts.get_contacts,
            limit=limit,
            filters=filters,
            dict_unpack_value='contacts',
        )

        if to_dictionary:
            return objects

        return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    def add_to_contact_group(self, contact_group: 'ContactGroup | str'):
        """Add to ContactGroup

        Args:
            contact_group: contact group to which add this contact
        """
        if isinstance(contact_group, str):
            contact_group = ContactGroup(self.connection, id=contact_group)

        contact_group.add_members([self])
        self.fetch()

    def remove_from_contact_group(self, contact_group: 'ContactGroup | str'):
        """Remove from ContactGroup

        Args:
            contact_group: contact group from which to remove this contact
        """
        if isinstance(contact_group, str):
            contact_group = ContactGroup(self.connection, id=contact_group)

        contact_group.remove_members([self])
        self.fetch()



"""This is the demo script to show how to manage translations. This script will
not work without replacing parameters with real values. Its basic goal is to
present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from io import StringIO
from mstrio.datasources import DatasourceInstance
from mstrio.object_management import list_translations, Translation
from mstrio.modeling.metric import Metric
from mstrio.server import Language

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

# Create connection based on workstation data
CONN = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
METRIC_ID = $metric_id
OBJECT_TYPE = $metric_type
LANG_LCID = $lang_lcid
LANG_2_LCID = $lang_2_lcid
LANG_ID = $lang_id
LANG_2_ID = $lang_2_id
LANGUAGE_OBJECT = $language_object
LANGUAGE_OBJECT_2 = $language_object_2
TARGET_ID = $target_id
TRANSLATION_VALUE = $translation_value
TRANSLATION_VALUE_ALTERED = $translation_value_altered
TRANSLATION_VALUE_2 = $translation_value_2
TRANSLATION_VALUE_2_ALTERED = $translation_value_2_altered

# You can manage translations in 2 ways, through the Translation class that
# allows managing Translations on any object.
# Or through the translation methods added to all objects that support
# translations, that are supported by mstrio-py.

# Below are snippets on using the Translation class methods:

# Listing translations for an object
list_translations(
    connection=CONN, id=METRIC_ID, object_type=OBJECT_TYPE, project_id=CONN.project_id
)

# You can also specify a list of languages for the listing function
# The function will automatically list translations for only those languages
# For the languages you have to provide a list.
# To specify a language in the list you need to provide either:
# - Language class objects
# - specific Language IDs
# - specific Language lcids
# The three types of language argument can be mixed in the list
list_translations(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    project_id=CONN.project_id,
    languages=[LANG_LCID, LANG_2_ID],
)

# Add translations to an Object

# The translations are specified with a list of helper classes
# Each helper class has to contain the target_language, value and target_id
# For target_language you can either specify:
# - Language class object
# - specific Language ID
# - specific Language lcid

Translation.add_translation(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    translations=[
        Translation.OperationData(
            target_language=LANG_LCID,
            value=TRANSLATION_VALUE,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_LCID,
            value=TRANSLATION_VALUE_2,
            target_id=TARGET_ID,
        ),
    ],
    project_id=CONN.project_id,
)

# Alter translations of an Object

# The translations are specified with a list of helper classes
# The helper classes have identical structure to the ones provided when adding
Translation.alter_translation(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    translations=[
        Translation.OperationData(
            target_language=LANG_ID,
            value=TRANSLATION_VALUE_ALTERED,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_ID,
            value=TRANSLATION_VALUE_2_ALTERED,
            target_id=TARGET_ID,
        ),
    ],
    project_id=CONN.project_id,
)

# Remove translations from an Object

# The translations are specified with a list of helper classes
# The helper class for removing does not need to contain the value
Translation.remove_translation(
    connection=CONN,
    id=METRIC_ID,
    object_type=OBJECT_TYPE,
    translations=[
        Translation.OperationData(target_language=LANGUAGE_OBJECT, target_id=TARGET_ID),
        Translation.OperationData(
            target_language=LANGUAGE_OBJECT_2, target_id=TARGET_ID
        ),
    ],
    project_id=CONN.project_id,
)

# Translations can be managed from an object as well
metric = Metric(connection=CONN, id=METRIC_ID)

# List translations of the object
metric.list_translations()

# You can specify a languages list for the object as well
metric.list_translations(languages=[LANG_LCID, LANG_2_ID])

# Add translations to the object
# Only the list of translation helper classes needs to be provided
metric.add_translation(
    translations=[
        Translation.OperationData(
            target_language=LANG_LCID,
            value=TRANSLATION_VALUE,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_LCID,
            value=TRANSLATION_VALUE_2,
            target_id=TARGET_ID,
        ),
    ],
)

# Alter translations of the object
# Only the list of translation helper classes needs to be provided
metric.alter_translation(
    translations=[
        Translation.OperationData(
            target_language=LANG_LCID,
            value=TRANSLATION_VALUE_ALTERED,
            target_id=TARGET_ID,
        ),
        Translation.OperationData(
            target_language=LANG_2_LCID,
            value=TRANSLATION_VALUE_2_ALTERED,
            target_id=TARGET_ID,
        ),
    ],
)

# Remove translations from the object
# Only the list of translation helper classes needs to be provided
metric.remove_translation(
    translations=[
        Translation.OperationData(target_language=LANG_LCID, target_id=TARGET_ID),
        Translation.OperationData(target_language=LANG_2_LCID, target_id=TARGET_ID),
    ],
)

# Bulk actions using CSV, JSON, SQL database or dataframes can also be
# performed on multiple objects at once

# To ensure correct import, the columns from the original file/database cannot
# be renamed. Columns can be deleted (with the exception of object ID,
# object type and target ID) but new columns cannot be added. The column
# order does not matter.

# Importing is supported from normalized file/database only

# Define variables which can be later used in a script
DATASOURCE_ID = $datasource_id
TABLE_NAME = $table_name
PATH_FOR_CSV = $path_for_csv
PATH_FOR_JSON = $path_for_json
SEPARATOR = $separator

# Exporting and importing from CSV
# If no path is specified, the CSV file will be returned as a string
CSV_LIST = Translation.to_csv_from_list(
    connection=CONN,
    object_list=[metric],
)

# If path is specified, CSV file will be created at the specified path
Translation.to_csv_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_CSV,
)

# Extra columns can be added by setting the following flags to True:
# add_object_creation_date
# add_object_description
# add_object_last_modified_date
# add_object_path
# add_object_version
Translation.to_csv_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_CSV,
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)

# Translations can be exported to CSV for specified languages only
Translation.to_csv_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_CSV,
    languages=[LANG_ID, LANG_2_ID],
)

# Translations can be exported to CSV in a denormalized format
# And with a specific separator
Translation.to_csv_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_CSV,
    denormalized_form=True,
    separator=SEPARATOR,
)

# Add translations from CSV
# If CSV is to be given as a string, it has to be converted to a StringIO type string first
CSV_LIST_IOSTRING = StringIO(CSV_LIST)
Translation.add_translations_from_csv(connection=CONN, file_path=CSV_LIST_IOSTRING)

# Instead a path to the CSV file can be specified as well
Translation.add_translations_from_csv(connection=CONN, file_path=PATH_FOR_CSV)

# A separator can be specified for the import file if a custom separator was used
Translation.add_translations_from_csv(
    connection=CONN, file_path=PATH_FOR_CSV, separator=SEPARATOR
)

# Extra flags can be enabled for the import function as well
# delete=True will make the import delete translations that are empty in the CSV
# automatch_target_ids will make the import try to match the target IDs
# that are not present in the Object on the server to ones present in the file
# based on the content of the Translation for the default project language

Translation.add_translations_from_csv(
    connection=CONN, file_path=PATH_FOR_CSV, delete=True, automatch_target_ids=True
)

# Exporting and importing from JSON
# If no path is specified, the JSON file will be returned as a string
JSON_LIST = Translation.to_json_from_list(
    connection=CONN,
    object_list=[metric],
)

# If path is specified, JSON file will be created at the specified path
Translation.to_json_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_JSON,
)

# Extra columns can be added by setting the following flags to True:
# add_object_creation_date
# add_object_description
# add_object_last_modified_date
# add_object_path
# add_object_version
Translation.to_json_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_JSON,
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)

# Translations can be exported to JSON for specified languages only
Translation.to_json_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_JSON,
    languages=[LANG_ID, LANG_2_ID],
)

# Translations can be exported to JSON in a denormalized format
Translation.to_json_from_list(
    connection=CONN,
    object_list=[metric],
    file_path=PATH_FOR_JSON,
    denormalized_form=True,
)

# Add translations from JSON
# If JSON is to be given as a string, it has to be converted to a StringIO type string first
JSON_LIST_IOSTRING = StringIO(JSON_LIST)
Translation.add_translations_from_json(connection=CONN, file_path=JSON_LIST_IOSTRING)

# Instead a path to the JSON file can be specified as well
Translation.add_translations_from_json(connection=CONN, file_path=PATH_FOR_JSON)

# Extra flags can be enabled for the import function as well
# delete=True will make the import delete translations that are empty in the JSON
# automatch_target_ids will make the import try to match the target IDs
# that are not present in the Object on the server to ones present in the file
# based on the content of the Translation for the default project language

Translation.add_translations_from_json(
    connection=CONN, file_path=PATH_FOR_JSON, delete=True, automatch_target_ids=True
)

# Exporting and importing from databases
# To interact with a database we need a DatasourceInstance object
DATASOURCE = DatasourceInstance(connection=CONN, id=DATASOURCE_ID)
Translation.to_database_from_list(
    connection=CONN,
    object_list=[metric],
    table_name=TABLE_NAME,
    datasource=DATASOURCE,
)

# By default the method will ask if you want to overwrite the table if it already exists
# To bypass having to answer that question the flag force can be set to True
Translation.to_database_from_list(
    connection=CONN,
    object_list=[metric],
    table_name=TABLE_NAME,
    datasource=DATASOURCE,
    force=True,
)

# Extra columns can be added by setting the following flags to True:
# add_object_creation_date
# add_object_description
# add_object_last_modified_date
# add_object_path
# add_object_version
Translation.to_database_from_list(
    connection=CONN,
    object_list=[metric],
    table_name=TABLE_NAME,
    datasource=DATASOURCE,
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)

# Translations can be exported to the database for specified languages only
Translation.to_database_from_list(
    connection=CONN,
    object_list=[metric],
    table_name=TABLE_NAME,
    datasource=DATASOURCE,
    languages=[LANG_ID, LANG_2_ID],
)

# Translations can be exported to the database in a denormalized format
Translation.to_database_from_list(
    connection=CONN,
    object_list=[metric],
    table_name=TABLE_NAME,
    datasource=DATASOURCE,
    denormalized_form=True,
)

# Add translations from database
Translation.add_translations_from_database(
    connection=CONN, table_name=TABLE_NAME, datasource=DATASOURCE
)

# Extra flags can be enabled for the import function as well
# delete=True will make the import delete translations that are empty in the database
# automatch_target_ids will make the import try to match the target IDs
# that are not present in the Object on the server to ones present in the file
# based on the content of the Translation for the default project language

Translation.add_translations_from_database(
    connection=CONN,
    table_name=TABLE_NAME,
    datasource=DATASOURCE,
    delete=True,
    automatch_target_ids=True,
)

# Exporting and importing from dataframes
DATAFRAME = Translation.to_dataframe_from_list(
    connection=CONN,
    object_list=[metric],
)

# Extra columns can be added by setting the following flags to True:
# add_object_creation_date
# add_object_description
# add_object_last_modified_date
# add_object_path
# add_object_version
Translation.to_dataframe_from_list(
    connection=CONN,
    object_list=[metric],
    add_object_creation_date=True,
    add_object_description=True,
    add_object_last_modified_date=True,
    add_object_path=True,
    add_object_version=True,
)

# Translations can be exported to the dataframe for specified languages only
Translation.to_dataframe_from_list(
    connection=CONN,
    object_list=[metric],
    languages=[LANG_ID, LANG_2_ID],
)

# Translations can be exported to the dataframe in a denormalized format
Translation.to_dataframe_from_list(
    connection=CONN,
    object_list=[metric],
    denormalized_form=True,
)

# Add translations from dataframe
Translation.add_translations_from_dataframe(connection=CONN, dataframe=DATAFRAME)

# Extra flags can be enabled for the import function as well
# delete=True will make the import delete translations that are empty in the dataframe
# automatch_target_ids will make the import try to match the target IDs
# that are not present in the Object on the server to ones present in the file
# based on the content of the Translation for the default project language

Translation.add_translations_from_dataframe(
    connection=CONN, dataframe=DATAFRAME, delete=True, automatch_target_ids=True
)



"""
    This is a template for creation of Datasource Script.

    Three methods need to be created for the workflow to work:
        browse(): retrieves the catalog information
        preview(): gets partial data for preview, data refine and schema change
        publish(): gets the data published and stores the data into a cube
    All three methods have to return data in the format of Pandas DataFrame.
"""

"""
    Simple Example:
    ```
    import pandas as pd

    # construct 2 tables to import
    year_cost_json = '[{"Year": 2010, "Cost": 7343097.05960049}, ' \\
            '{"Year": 2011, "Cost": 9777520.64980007}, ' \\
            '{"Year": 2012, "Cost": 12609466.8109989}]'
    df_cost = pd.json_normalize(pd.io.json.loads(year_cost_json))
    year_revenue_json = '[{"Year": 2010, "Revenue": 8647238.10000023}, ' \\
            '{"Year": 2011, "Revenue": 11517606}, ' \\
            '{"Year": 2012, "Revenue": 14858864.0500001}]'
    df_revenue = pd.json_normalize(pd.io.json.loads(year_revenue_json))
    all_tables = dict({'YearlyCost': df_cost, 'YearlyRevenue': df_revenue})

    def browse():
        return all_tables

    def preview(table_name, row_count):
        if table_name not in all_tables:
            raise ValueError(f"Table {table_name} is not found.")
        return all_tables.get(table_name).head(row_count)

    def publish(table_name):
        if table_name not in all_tables:
            raise ValueError(f"Table {table_name} is not found.")
        return all_tables.get(table_name)
    ```
"""


def browse():
    """Retrieve the catalog information.
    Returns:
        a dict object. The keys of the dict should be table names of the Python
        data source, and the values are normalized in Pandas DataFrame format.
        Each DataFrame value will contain a table's column infos.
    """
    pass


def preview(table_name, row_count):
    """Get partial data for preview, data refine and schema change.
    Args:
        table_name: name of the selected table.
        row_count: number of rows to be retrieved.
    Returns:
        a Pandas DataFrame object with `row_count` rows.
    """
    pass


def publish(table_name):
    """Get the data published and store the data into the cube
    Args:
        table_name: name of the selected table.
    Returns:
        a Pandas DataFrame object with all data.
    """
    pass



import csv
from mstrio.connection import get_connection
from mstrio.helpers import AggregatedRights
from mstrio.modeling.security_filter.security_filter import list_security_filters
from mstrio.users_and_groups.user import User, create_users_from_csv
from mstrio.users_and_groups.user_group import UserGroup, list_user_groups

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# set correct user group access in tree view for user groups
# assumtions about CSV - csv is created from User.to_csv_from_list with at least
# two columns: username and full_name
user_list = create_users_from_csv(conn, 'users.csv')
user_group = list_user_groups(conn)[0]
user_group.acl_add(
    right=AggregatedRights.VIEW,
    trustees=user_list,
    denied=False,
    inheritable=True,
    propagate_to_children=True,
)

# reset the password for those users and check-in the option
# "User must change password at the next login"
for user in user_list:
    user.alter(password='new_password', require_new_password=True)

# import user group structure from CSV file
# assumptions - you have a prepared list CSV of two columns:
# user_name , user_group_name with headers added
with open('tree_structure.csv', encoding='utf-8') as structure:
    user_rows = csv.DictReader(structure)
    for row in user_rows:
        user_name = row['user_name']
        user_group_name = row['user_group_name']
        user = User(conn, name=user_name)
        user_group = UserGroup(conn, name=user_group_name)
        user_group.add_users(user)

# assigning correct security filter (the same name) for the user group
user_group = UserGroup(conn, name='AI Users')
security_filter = list_security_filters(conn, name='security_filter')[0]
user_group.apply_security_filter(security_filter)



import datetime as dt
from collections import defaultdict

import pandas as pd


class Model:
    """Internal utility for generating the definition of multi-table and
    single-table datasets.

    Create the definition of a super cube containing one or more tables. The
    definition includes the name and description of the super cube and the name
    and description of each table, attribute, and metric within the super cube.

    Attributes:
    """

    # dictionary key names
    _KEY_TABLE_NAME = 'table_name'
    _KEY_DATA_FRAME = 'data_frame'
    _KEY_AS_ATTR = 'to_attribute'
    _KEY_AS_METR = 'to_metric'
    _KEY_UPDATE_POL = 'update_policy'
    _KEY_TABLES = [_KEY_TABLE_NAME, _KEY_DATA_FRAME, _KEY_AS_ATTR, _KEY_AS_METR]

    _MAX_DESC_LEN = 250

    _INVALID_CHARS = [
        '\\',
        '"',
        '[',
        ']',
    ]  # check for invalid characters in column names

    def __init__(
        self,
        tables,
        name,
        description=None,
        folder_id=None,
        ignore_special_chars=False,
        attr_forms_mapping=None,
    ):
        """Initializes Model with tables, a name, and an optional description.

        Args:
            tables (list of dicts): List of one or more dictionaries with keys
                `table_name`, `data_frame`, and optionally `as_attribute` and
                `as_metric`. Note that `as_attribute` and `as_metric` should be
                used when the default Python data type (e.g. `int`) should be
                converted to an attribute instead of a metric in MicroStrategy.
            name (str): Name of the data set.
            description (str, optional): Description of the data set. Must be
                less than or equal to 250 characters.
            folder_id (str, optional): ID of the shared folder that the super
                cube should be created within. If `None`, defaults to the user's
                My Reports folder.
        """

        self._ignore_special_chars = ignore_special_chars

        # check integrity of tables list
        self._check_table_list(tables=tables)

        # check super cube name params
        self._name = name
        self._check_param_str(self._name, msg="SuperCube name should be a string.")
        self._check_param_len(
            self._name,
            msg=f"SuperCube name should be <= {self._MAX_DESC_LEN} characters.",
            max_length=self._MAX_DESC_LEN,
        )
        self._check_param_inv_chars(
            self._name,
            msg="SuperCube name cannot contain '{}', '{}', '{}', '{}'.".format(
                *self._INVALID_CHARS
            ),
            invalid_chars=self._INVALID_CHARS,
        )

        # check super cube description params
        if description is None:
            self._description = ""
        else:
            self._description = description
            self._check_param_str(
                self._description, msg="SuperCube description should be a string."
            )
            self._check_param_len(
                self._description,
                msg=(
                    f"SuperCube description should be <= {self._MAX_DESC_LEN}"
                    " characters."
                ),
                max_length=self._MAX_DESC_LEN,
            )

        # check folder_id param
        self._folder_id = folder_id or ''

        # init lists to accumulate table, attr, metric definitions and model
        self._tables = []
        self._attributes = []
        self._metrics = []
        self._model = None

        forms_mapping = attr_forms_mapping or []
        # build the model
        self._build(tables=tables, forms_mapping=forms_mapping)

    def get_model(self):
        """Return the model object."""
        return self._model

    def _build(self, tables, forms_mapping):
        """Generates the data model by mapping attributes and metrics from list
        of tables."""

        table2columns_used = self._add_user_defined_attribute_forms(forms_mapping)

        # Map tables one by one
        for table in tables:
            mapped_columns = table2columns_used[table['table_name']]
            self._map_table(table, mapped_columns)

        # set model object
        self._model = {
            'name': self._name,
            'description': self._description,
            'folderId': self._folder_id,
            'tables': self._tables,
            'metrics': self._metrics,
            'attributes': self._attributes,
        }

    def _add_user_defined_attribute_forms(self, forms_mapping):
        table2columns_used = defaultdict(set)

        for attr in forms_mapping:
            self._attributes.append(attr.to_dict())
            for form in attr.forms:
                for expr in form.expressions:
                    table2columns_used[expr.table].add(expr.column)

        return table2columns_used

    def _map_table(self, table, skip_columns):
        df = table[self._KEY_DATA_FRAME]

        # map column names and column types
        col_names = self._get_col_names(df)
        col_types = self._get_col_types(df)
        table_name = table[self._KEY_TABLE_NAME]

        # map tables
        self._add_table(name=table_name, col_names=col_names, col_types=col_types)

        # map attributes and metrics
        for name, column_type in zip(col_names, col_types):
            if name in skip_columns:
                continue

            if name in table.get(self._KEY_AS_ATTR, []):
                self._add_attribute(name, table_name)
            elif name in table.get(self._KEY_AS_METR, []) or self._is_metric(
                column_type
            ):
                self._add_metric(name, table_name)
            else:
                self._add_attribute(name, table_name)

    def _get_col_types(self, table):
        """Map column types from each column in the list of table."""
        list_dtypes_values = list(table.dtypes.values)

        for i, column_name in enumerate(table.columns):
            if list_dtypes_values[i] == 'object':
                column_element = table[column_name][0]

                if isinstance(column_element, dt.time):
                    list_dtypes_values[i] = 'time'
                if isinstance(column_element, dt.date):
                    list_dtypes_values[i] = 'date'

        return [self._map_data_type(datatype) for datatype in list_dtypes_values]

    def _add_metric(self, name, table_name):
        """Add a metric to a metric list instance."""
        self._metrics.append(
            {
                'name': name,
                'expressions': [{'tableName': table_name, 'columnName': name}],
            }
        )

    def _add_attribute(self, name, table_name):
        """Add an attribute to an attribute list instance."""
        self._attributes.append(
            {
                'name': name,
                'attributeForms': [
                    {
                        'category': 'ID',
                        'expressions': [{'tableName': table_name, 'columnName': name}],
                    }
                ],
            }
        )

    def _add_table(self, name, col_names, col_types):
        """Add a table to a table list instance."""
        self._tables.append(
            {
                'name': name,
                'columnHeaders': [
                    {'name': name, 'dataType': typ}
                    for name, typ in zip(col_names, col_types)
                ],
            }
        )

    def _check_table_list(self, tables):
        """Check integrity of table list parameter."""

        # tables must be a list
        if not isinstance(tables, list):
            raise TypeError("Elements of tables must be a list of dicts.")

        # tables cannot be length 0
        if len(tables) == 0:
            raise ValueError("No tables have been added to the super cube.")

        # check integrity of each table passed to tables
        for table in tables:
            self._check_table(table)

    def _check_table(self, table):
        """Check integrity of table parameter."""

        # force all list elements to be a dict with specific names
        table_keys_list = "', '".join(self._KEY_TABLES)
        msg = f"Each table must be a dictionary with keys: '{table_keys_list}'."

        if not isinstance(table, dict):
            raise TypeError(msg)

        if not all(k in table for k in (self._KEY_TABLE_NAME, self._KEY_DATA_FRAME)):
            raise ValueError(msg)

        # check that the value of the data frame key is a pandas data frame
        if not isinstance(table[self._KEY_DATA_FRAME], pd.DataFrame):
            msg = (
                "Pandas DataFrame must be passed as the value in the "
                f"'{self._KEY_DATA_FRAME}' key."
            )
            raise TypeError(msg)

        # check for presence of invalid characters in data frame column names
        if not self._ignore_special_chars and any(
            col
            for col in table[self._KEY_DATA_FRAME].columns
            for inv in self._INVALID_CHARS
            if inv in col
        ):
            msg = "Column names cannot contain '{}', '{}', '{}', '{}'".format(
                *self._INVALID_CHARS
            )
            raise ValueError(msg)

    @staticmethod
    def _get_col_names(table):
        """Returns column names from a table as a list."""
        return list(table.columns)

    @staticmethod
    def _map_data_type(datatype):
        """Maps a Python data type to a MicroStrategy data type."""
        return {
            'object': 'STRING',
            'int32': 'INTEGER',
            'int64': 'BIGINTEGER',
            'float32': 'DOUBLE',
            'float64': 'DOUBLE',
            'bool': 'BOOLEAN',
            'datetime64[ns]': 'DATETIME',
            'date': 'DATE',
            'time': 'TIME',
        }.get(str(datatype))

    @staticmethod
    def _check_param_len(param, msg, max_length):
        if len(param) > max_length:
            raise ValueError(msg)

        return True

    @staticmethod
    def _check_param_str(param, msg):
        if not isinstance(param, str):
            raise TypeError(msg)

        return True

    @staticmethod
    def _check_param_inv_chars(param, msg, invalid_chars):
        if any(inv for inv in invalid_chars if inv in param):
            raise ValueError(msg)

    @staticmethod
    def _is_metric(datatype):
        """Helper function for determining if the requested datatype is (by
        default) a metric or attribute."""
        return datatype in ['DOUBLE', 'INTEGER', 'BIGINTEGER', 'BIGDECIMAL']

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def folder_id(self):
        return self._folder_id

    @property
    def tables(self):
        return self._tables

    @property
    def attributes(self):
        return self._attributes

    @property
    def metrics(self):
        return self._metrics



"""This is the demo script to show how administrator can manage schedules
and events.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.distribution_services import (
    list_schedules, list_subscriptions, Schedule, ScheduleEnums, ScheduleTime, Subscription
)
from mstrio.distribution_services import Event, list_events
from mstrio.connection import get_connection
from mstrio.utils.wip import module_wip, WipLevels

# Currently there is no way to delete schedules with a script in Workstation.
# The functionality can be used in standalone Python runtime.
module_wip(globals(), level=WipLevels.WARNING)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# get list of all schedules
all_schedules = list_schedules(connection=conn)

# get all event based schedules
event_based_schedules = list_schedules(
    connection=conn, schedule_type=Schedule.ScheduleType.EVENT_BASED.value
)
# see distribution_services/device/schedule.py for ScheduleType values

# get all expired Schedules
expired_schedules = list_schedules(connection=conn, expired=True)

# get all events
events = list_events(connection=conn)

# Define variables which can be later used in a script
SCHEDULE_ID = $schedule_id  # id for Schedule object lookup
SCHEDULE_NAME = $schedule_name  # name for Schedule object

# get schedule by id
schedule = Schedule(connection=conn, id=SCHEDULE_ID)
# get schedule by name
schedule = Schedule(connection=conn, name=SCHEDULE_NAME)
# list schedule's properties
schedule.list_properties()

# Define variables which can be later used in a script
SCHEDULE_EXPIRE_DATE = $schedule_expire_date  # YYYY-MM-DD format
SCHEDULE_DESCRIPTION = $schedule_description

# disable/enable (pass new schedule expire time)
schedule.disable()
schedule.enable(SCHEDULE_EXPIRE_DATE)

# alter schedule properties
schedule.alter(description=SCHEDULE_DESCRIPTION)

# Define a variable which can be later used in a script
SCHEDULE_EXECUTE_TIME = $schedule_execute_time  # HH:MM:SS format

SCHEDULE_TIME = ScheduleTime.from_details(
    recurrence_pattern=ScheduleEnums.RecurrencePattern.MONTHLY,
    monthly_pattern=ScheduleEnums.MonthlyPattern.DAYS_OF_MONTH,
    days_of_month=['1', '3', '10', '28'],
    repeat_interval=2,
    execution_pattern=ScheduleEnums.ExecutionPattern.ONCE,
    execution_time=SCHEDULE_EXECUTE_TIME
)
# see distribution_services/schedule/schedule_time.py for RecurrencePattern,
# MonthlyPattern and ExecutionPattern values

# altering schedule type is also possible;
# changing event based schedule to time based (using ScheduleTime class):
schedule.alter(time=SCHEDULE_TIME)

# Define variables which can be later used in a script
SCHEDULE_START_DATE = $schedule_start_date  # YYYY-MM-DD format
SCHEDULE_STOP_DATE = $schedule_stop_date  # YYYY-MM-DD format

# Create time based schedule. Note that some parameters refer to particular time patterns and
# depend on setting recurrence_pattern or execution_pattern to respective values. In this case,
# setting daily_pattern to DAY requires specifying repeat_interval, and setting execution_pattern
# to ONCE requires specifying execution_time.
new_schedule = Schedule.create(
    connection=conn,
    name=SCHEDULE_NAME,
    schedule_type=Schedule.ScheduleType.TIME_BASED,
    start_date=SCHEDULE_START_DATE,
    recurrence_pattern=ScheduleEnums.RecurrencePattern.DAILY,
    daily_pattern=ScheduleEnums.DailyPattern.DAY,
    repeat_interval=2,
    execution_pattern=ScheduleEnums.ExecutionPattern.ONCE,
    execution_time=SCHEDULE_EXECUTE_TIME,
)
# see distribution_services/device/schedule.py for ScheduleType values
# see distribution_services/schedule/schedule_time.py for RecurrencePattern,
# DailyPattern and ExecutionPattern values

# Define a variable which can be later used in a script
SCHEDULE_EVENT_NAME = $schedule_event_name  # name of existing event to base a schedule on

# create event based schedule
event = Event(connection=conn, name=SCHEDULE_EVENT_NAME)
new_schedule = Schedule.create(
    connection=conn,
    name=SCHEDULE_NAME,
    schedule_type=Schedule.ScheduleType.EVENT_BASED,
    event_id=event.id,
    start_date=SCHEDULE_START_DATE,
    stop_date=SCHEDULE_STOP_DATE
)
# see distribution_services/device/schedule.py for ScheduleType values

# delete schedule
# (Note: deleting schedule will also delete related subscriptions!)
schedule.delete()

# get a list of subscriptions using schedule under the given project
all_subscriptions = list_subscriptions(conn, project_name=PROJECT_NAME)
subs_with_schedule = [sub for sub in all_subscriptions if sub.schedules[0].id == schedule.id]

# Define a variable which can be later used in a script
SUBSCRIPTION_ID = $subscription_id  # id for Subscription object lookup

# good practice is to check if subscription is using the schedule,
# which we want to delete and if so alter it before deletion
other_schedule = Schedule(connection=conn, name=SCHEDULE_NAME)
sub1 = Subscription(connection=conn, subscription_id=SUBSCRIPTION_ID, project_name=PROJECT_NAME)
sub1.alter(schedules=other_schedule)

# delete all expired schedules - user will be prompted to confirm every deletion
# (Note: deleting schedule will also delete related subscriptions!)
expired_schedules = list_schedules(connection=conn, expired=True)
for sch in expired_schedules:
    sch.delete()

# changing schedules for multiple subscriptions can be easily done using simple
# for loop (batch execution substitute)
sub2 = Subscription(connection=conn, subscription_id=SUBSCRIPTION_ID, project_name=PROJECT_NAME)
list_of_subscriptions = [sub1, sub2]
for subscription in list_of_subscriptions:
    subscription.alter(schedules=schedule)



import time
from mstrio.connection import get_connection
from mstrio.project_objects.datasets.cube import CubeStates
from mstrio.project_objects.datasets import OlapCube
from mstrio.distribution_services.subscription.subscription_manager import (
    list_subscriptions,
)


conn = get_connection(workstationData, 'MicroStrategy Tutorial')


# Select an intelligent cube and refresh it
example_cube = OlapCube(
    connection=conn, id="8CCD8D9D4051A4C533C719A6590DEED8"
)  # Intelligent Cube - Drilling
example_cube.publish()


# Wait for cube refresh to complete successfully.
example_cube.refresh_status()
while "Processing" in CubeStates.show_status(example_cube.status):
    time.sleep(1)
    example_cube.refresh_status()


# Search for subscriptions dependent on that cube
# (find dashboards dependent on cube, and then subscriptions
# dependent on those dashboards found)
dependent_dashboards = example_cube.list_dependents()
all_subscriptions = list_subscriptions(conn)
subscriptions_to_send = []
for d in dependent_dashboards:
    # For each found Dashboard, find all Subscriptions using it.
    # This step requires filtering a list of Subscriptions with Python
    # list operations, as Document/Dashboard query responses do not backlink
    # to Subscriptions.
    dashboard_id = d['id']

    subscriptions_to_send += [
        sub
        for sub in all_subscriptions
        if any(content.id == dashboard_id for content in sub.contents)
    ]

# Trigger (run) the subscriptions
for s in subscriptions_to_send:
    s.execute()



"""This is the demo script to show how to manage datasources.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.datasources import (
    DatasourceConnection,
    DatasourceInstance,
    DatasourceLogin,
    DatasourceMap,
    DatasourceType,
    Dbms,
    Driver,
    ExecutionMode,
    Gateway,
    list_available_dbms,
    list_datasource_connections,
    list_datasource_instances,
    list_datasource_logins,
    list_datasource_mappings,
    list_drivers,
    list_gateways,
)
from mstrio.connection import get_connection
from mstrio.server import list_languages, Language

# Define variables which can be later used in a script
PROJECT_ID = $project_id  # Project ID to connect to
PROJECT_NAME = $project_name  # Insert project name to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Manage datasource logins
# Get a list of datasource logins
datasource_login_list = list_datasource_logins(connection=conn)
print(datasource_login_list)

# Define variables which can be later used in a script
# Insert ID for login to datasource here
DATASOURCE_LOGIN_ID = $datasource_login_id
# Insert name for login to datasource here
DATASOURCE_LOGIN_NAME = $datasource_login_name

# Get datasource login by id
datasource_login = DatasourceLogin(conn, id=DATASOURCE_LOGIN_ID)
print(datasource_login)

# Get datasource login by name
datasource_login = DatasourceLogin(conn, name=DATASOURCE_LOGIN_NAME)
print(datasource_login)

# Define variables which can be later used in a script
# Insert a new name for new login to datasource here
NEW_DATASOURCE_LOGIN_NAME = $new_datasource_login_name
# Insert a new name of user in Datasource here
NEW_DATASOURCE_USERNAME = $new_datasource_username
# Insert a new datasource password here
NEW_DATASOURCE_PASSWORD = $new_datasource_password
# Insert a new datasource login description
NEW_DATASOURCE_LOGIN_DESCRIPTION = $new_datasource_login_description

# Create new datasource login
datasource_login = DatasourceLogin.create(
    connection=conn,
    name=NEW_DATASOURCE_LOGIN_NAME,
    username=NEW_DATASOURCE_USERNAME,
    password=NEW_DATASOURCE_PASSWORD,
    description=NEW_DATASOURCE_LOGIN_DESCRIPTION,
)
print(datasource_login)

# Define variables which can be later used in a script
# Insert new name for login to datasource here
DATASOURCE_LOGIN_NEW_NAME = $datasource_login_new_name
# Insert new datasource login description
DATASOURCE_LOGIN_NEW_DESCRIPTION = $datasource_login_new_description

# Update a datasource login
datasource_login.alter(
    name=DATASOURCE_LOGIN_NEW_NAME,
    description=DATASOURCE_LOGIN_NEW_DESCRIPTION,
)
print(datasource_login)

# List properties of a datasource login
print(datasource_login.list_properties())

# Delete a datasource login
datasource_login.delete(force=True)

# Manage datasource connections
# List all datasource connections
datasource_conn_list = list_datasource_connections(connection=conn)
print(datasource_conn_list)

# Define variables which can be later used in a script
# Insert ID of datasource connection here
DATASOURCE_CONNECTION_ID = $datasource_connection_id
# Insert name of datasource connection here
DATASOURCE_CONNECTION_NAME = $datasource_connection_name

# Get datasource connection by id
datasource_connection = DatasourceConnection(conn, id=DATASOURCE_CONNECTION_ID)
print(datasource_connection)

# Get datasource connection by name
datasource_connection = DatasourceConnection(conn, name=DATASOURCE_CONNECTION_NAME)
print(datasource_connection)

# Define variables which can be later used in a script
# Insert new name of datasource connection here
NEW_DATASOURCE_CONNECTION_NAME = $new_datasource_connection_name
# Insert new description of datasource connection here
NEW_DATASOURCE_CONNECTION_DESCRIPTION = $new_datasource_connection_description

# Create a new datasource connection
datasource_connection = DatasourceConnection.create(
    connection=conn,
    name=NEW_DATASOURCE_CONNECTION_NAME,
    description=NEW_DATASOURCE_CONNECTION_DESCRIPTION,
    # the ExecutionMode values can be found in datasources/datasource_connection.py
    execution_mode=ExecutionMode.SYNCHRONOUS,
    datasource_login=DATASOURCE_LOGIN_ID,
)

# Define variables which can be later used in a script
# Insert new name of datasource connection
DATASOURCE_CONNECTION_NEW_NAME = $datasource_connection_new_name
# Insert new description of datasource connection
DATASOURCE_CONNECTION_NEW_DESCRIPTION = $datasource_connection_new_description

# Update a datasource connection
datasource_connection.alter(
    name=DATASOURCE_CONNECTION_NEW_NAME,
    description=DATASOURCE_CONNECTION_NEW_DESCRIPTION,
)
print(datasource_connection)

# List properties of a datasource connection
print(datasource_connection.list_properties())

# Delete a datasource connection
datasource_connection.delete(force=True)


# Note that conversion to DSN-less is supported on certified gateways:
# Amazon Redshift; Azure Synapse Analytics; ExasolI;
# BM Db2 for Linux, UNIX and Windows; MySQL; Oracle; PostgreSQL;
# Salesforce; Cloudera Hive; Cloudera Impala; Google BigQuery; Spark SQL;
# SAP HANA; Snowflake; SQL Server; Teradata

# Define variables which can be later used in a script
# Insert ID of datasource connection that had DSN connection string
DATASOURCE_CONNECTION_ID_DSN = $datasource_connection_id_dsn

# Initialize datasource connection object
ds_conn = DatasourceConnection(conn, id=DATASOURCE_CONNECTION_ID_DSN)

# List datasource connection properites before conversion
print(ds_conn.list_properties())

# This method is supported from Update 9 server version
# Convert datasource connection to DSN-less
ds_conn.convert_to_dsn_less()

# List datasource connection properites after conversion
print(ds_conn.list_properties())

# Define variables which can be later used in a script
# Insert IDs for datasource connections here
DATASOURCE_CONNECTION_ID_DSN_LIST = $datasource_connection_id_dsn_list

# Convert datasource connections one by one from list
for ds_conn_id in DATASOURCE_CONNECTION_ID_DSN_LIST:
    ds_conn = DatasourceConnection(conn, id=ds_conn_id)
    # This method is supported from Update 9 server version
    ds_conn.convert_to_dsn_less()


# Manage datasource instances
# List all datasources
datasources = list_datasource_instances(connection=conn)
print(datasources)

# List all datasources by project
datasources = list_datasource_instances(connection=conn, project=PROJECT_ID)
print(datasources)

# List all datasources by datasource connection
datasources = list_datasource_instances(
    connection=conn, datasource_connection={'id': DATASOURCE_CONNECTION_ID}
)
print(datasources)

# Define variables which can be later used in a script
# Insert ID for datasource instance here
DATASOURCE_INSTANCE_ID = $datasource_instance_id
# Insert name for datasource instance here
DATASOURCE_INSTANCE_NAME = $datasource_instance_name

# Get datasource instance by id
datasource_instance = DatasourceInstance(conn, id=DATASOURCE_INSTANCE_ID)
print(datasource_instance)

# Get datasource instance by name
datasource_instance = DatasourceInstance(conn, name=DATASOURCE_INSTANCE_NAME)
print(datasource_instance)

# List dbms
dbms = list_available_dbms(connection=conn)
print(dbms)

# Define variables which can be later used in a script
DBMS_ID = $dbms_id  # Insert ID of DBMS that you want to find
DBMS_NAME = $dbms_name  # Insert name of DBMS that you want to find

# Get a DBMS by id
dbms = Dbms(conn, id=DBMS_ID)
print(dbms)

# Get a DBMS by name
dbms = Dbms(conn, name=DBMS_NAME)
print(dbms)

# Define variables which can be later used in a script
# Insert a new name for datasource instance here
NEW_DATASOURCE_INSTANCE_NAME = $new_datasource_instance_name
# Insert a new description for datasource instance here
NEW_DATASOURCE_INSTANCE_DESCRIPTION = $new_datasource_instance_description
# Insert a new table prefix for datasource instance here
NEW_DATASOURCE_INSTANCE_TABLE_PREFIX = $new_datasource_instance_table_prefix

# Create a datasource instance
datasource_instance = DatasourceInstance.create(
    connection=conn,
    name=NEW_DATASOURCE_INSTANCE_NAME,
    description=NEW_DATASOURCE_INSTANCE_DESCRIPTION,
    dbms=DBMS_ID,
    datasource_connection=DATASOURCE_CONNECTION_ID,
    table_prefix=NEW_DATASOURCE_INSTANCE_TABLE_PREFIX,
    # The DatasourceType values can be found in datasources/datasource_instance.py
    datasource_type=DatasourceType.RESERVED
)
print(datasource_instance)

# Define variables which can be later used in a script
# Insert new name for edited datasource instance here
DATASOURCE_INSTANCE_NEW_NAME = $datasource_instance_new_name
# Insert new description for edited datasource instance here
DATASOURCE_INSTANCE_NEW_DESCRIPTION = $datasource_instance_new_description
# Insert new table prefix for edited datasource instance here
DATASOURCE_INSTANCE_TABLE_NEW_PREFIX = $datasource_instance_table_new_prefix

# Update a datasource instance
datasource_instance.alter(
    name=DATASOURCE_INSTANCE_NEW_NAME,
    description=DATASOURCE_INSTANCE_NEW_DESCRIPTION,
    table_prefix=DATASOURCE_INSTANCE_TABLE_NEW_PREFIX
)
print(datasource_instance)

# List properties of a datasource instance
print(datasource_instance.list_properties())

# Delete a datasource instance
datasource_instance.delete(force=True)


# Note that conversion to DSN-less is supported on certified gateways:
# Amazon Redshift; Azure Synapse Analytics; ExasolI;
# BM Db2 for Linux, UNIX and Windows; MySQL; Oracle; PostgreSQL;
# Salesforce; Cloudera Hive; Cloudera Impala; Google BigQuery; Spark SQL;
# SAP HANA; Snowflake; SQL Server; Teradata

# Define variables which can be later used in a script
# Insert ID for datasource instance here
DATASOURCE_INSTANCE_ID_DSN = $datasource_instance_id_dsn

# Get datasource instance by id
ds_instance = DatasourceInstance(conn, id=DATASOURCE_INSTANCE_ID_DSN)

# This method is supported from Update 9 server version
# Convert datasource embedded connection from DSN to DSN-less format
ds_instance.convert_ds_connection_to_dsn_less()

# Define variables which can be later used in a script
# Insert IDs for datasource instances here
DATASOURCE_INSTANCE_ID_DSN_LIST = $datasource_instance_id_dsn_list

# Convert datasource instance embedded connection one by one from list
for ds_id in DATASOURCE_INSTANCE_ID_DSN_LIST:
    ds_instance = DatasourceInstance(conn, id=ds_id)
    # This method is supported from Update 9 server version
    ds_instance.convert_ds_connection_to_dsn_less()


# Manage connection mappings

# List all locales
language_list = list_languages(connection=conn)
print(language_list)

# Define variables which can be later used in a script
LANGUAGE_ID = $language_id
LANGUAGE_NAME = $language_name

# Initialize language object by id
language = Language(conn, id=LANGUAGE_ID)
print(language)

# Initialize language object by name
language = Language(conn, name=LANGUAGE_NAME)
print(language)

# List all connection mappings
connection_mapping_list = list_datasource_mappings(connection=conn)
print(connection_mapping_list)

# Define a variable which can be later used in a script
USER_OR_USER_GROUP_ID = $user_or_user_group_id  # Insert ID of a user or user group
                                                # that you wish to be mapped
                                                # with a connection mapping

# List connection mappings filtered by user
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    user=USER_OR_USER_GROUP_ID,
)
print(connection_mapping_list)

# List connection mappings filtered by datasource connection
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    ds_connection=DATASOURCE_CONNECTION_ID,
)
print(connection_mapping_list)

# List connection mappings filtered by datasource instance
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    datasource=DATASOURCE_INSTANCE_ID,
)
print(connection_mapping_list)

# List connection mappings filtered by datasource login
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    login=DATASOURCE_LOGIN_ID,
)
print(connection_mapping_list)

# List connection mappings filtered by language ID
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    locale=LANGUAGE_ID,
)
print(connection_mapping_list)

# List connection mappings filtered by language name
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    locale=LANGUAGE_NAME,
)
print(connection_mapping_list)

# List default connection mappings for a project
connection_mapping_list = list_datasource_mappings(
    connection=conn,
    default_connection_map=True,
    project=PROJECT_ID,
)
print(connection_mapping_list)

# Define a variable which can be later used in a script
CONNECTION_MAPPING_ID = $connection_mapping_id  # Insert ID of connection mapping

# Initialise a connection mapping
connection_mapping = DatasourceMap(connection=conn, id=CONNECTION_MAPPING_ID)
print(connection_mapping)

# Initialise a default connection mapping by id for a project
connection_mapping = DatasourceMap(
    connection=conn,
    id=CONNECTION_MAPPING_ID,
    default_connection_map=True,
    project=PROJECT_ID,
)
print(connection_mapping)

# Initialise a default connection mapping for a project if only one default
# connection mapping exists for a project
connection_mapping = DatasourceMap(
    connection=conn,
    default_connection_map=True,
    project=PROJECT_ID,
)
print(connection_mapping)

# Create a connection mapping with default locale
connection_mapping = DatasourceMap.create(
    connection=conn,
    project=PROJECT_ID,
    user=USER_OR_USER_GROUP_ID,
    ds_connection=DATASOURCE_CONNECTION_ID,
    datasource=DATASOURCE_INSTANCE_ID,
    login=DATASOURCE_LOGIN_ID,
)
print(connection_mapping)

# Create a connection mapping with locale's id
connection_mapping = DatasourceMap.create(
    connection=conn,
    project=PROJECT_ID,
    user=USER_OR_USER_GROUP_ID,
    ds_connection=DATASOURCE_CONNECTION_ID,
    datasource=DATASOURCE_INSTANCE_ID,
    login=DATASOURCE_LOGIN_ID,
    locale=LANGUAGE_ID
)
print(connection_mapping)

# Create a connection mapping with locale's name
connection_mapping = DatasourceMap.create(
    connection=conn,
    project=PROJECT_ID,
    user=USER_OR_USER_GROUP_ID,
    ds_connection=DATASOURCE_CONNECTION_ID,
    datasource=DATASOURCE_INSTANCE_ID,
    login=DATASOURCE_LOGIN_ID,
    locale=LANGUAGE_NAME
)
print(connection_mapping)

# Define variables which can be later used in a script
# Insert new user or usergroup ID for connection mapping
CONNECTION_MAPPING_NEW_USER_OR_USER_GROUP = $connection_mapping_new_user_or_user_group
# Insert new login ID for connection mapping
CONNECTION_MAPPING_NEW_LOGIN = $connection_mapping_new_login

# Alter connection mapping
# NOTE: Altering connection mapping will change its ID
connection_mapping.alter(
    user=CONNECTION_MAPPING_NEW_USER_OR_USER_GROUP,
    login=CONNECTION_MAPPING_NEW_LOGIN,
)
print(connection_mapping)

# Alter connection mapping's locale by id
connection_mapping.alter(locale=LANGUAGE_ID)
print(connection_mapping)

# Alter connection mapping's locale by name
connection_mapping.alter(locale=LANGUAGE_NAME)
print(connection_mapping)

# Delete a connection mapping
connection_mapping.delete(force=True)


# Manage Drivers

# List all drivers
drivers = list_drivers(conn)
print(drivers)

# Define variables which can be later used in a script
DRIVER_ID = $driver_id

# List drivers by id
drivers = list_drivers(conn, id=DRIVER_ID)
print(drivers)

# Define variables which can be later used in a script
DRIVER_NAME = $driver_name

# List drivers by name
drivers = list_drivers(conn, name=DRIVER_NAME)
print(drivers)

# List enabled drivers
drivers = list_drivers(conn, is_enabled=True)
print(drivers)

# List disabled drivers
drivers = list_drivers(conn, is_enabled=False)
print(drivers)

# List ODBC drivers
drivers = list_drivers(conn, is_odbc=True)
print(drivers)

# List non-ODBC drivers
drivers = list_drivers(conn, is_odbc=False)
print(drivers)

# Initialize driver by id
driver = Driver(conn, id=DRIVER_ID)
print(driver)

# Initialize driver by name
driver = Driver(conn, name=DRIVER_NAME)
print(driver)

# Enable driver
driver.enable()

# Disable driver
driver.disable()

# Alter driver
driver.alter(is_enabled=True)


# Managing Gateways

# List all gateways
gateways = list_gateways(conn)
print(gateways)

# Define variables which can be later used in a script
GATEWAY_ID = $gateway_id

# List gateways by id
gateways = list_gateways(conn, id=GATEWAY_ID)
print(gateways)

# Define variables which can be later used in a script
GATEWAY_NAME = $gateway_name

# List gateways by name
gateways = list_gateways(conn, name=GATEWAY_NAME)
print(gateways)

# Define variables which can be later used in a script
GATEWAY_TYPE = $gateway_type

# List gateways by gateway type
gateways = list_gateways(conn, gateway_type=GATEWAY_TYPE)
print(gateways)

# Define variables which can be later used in a script
DB_TYPE = $db_type

# List gateways by database type
gateways = list_gateways(conn, db_type=DB_TYPE)
print(gateways)

# List certified gateways
gateways = list_gateways(conn, is_certified=True)
print(gateways)

# List not certified gateways
gateways = list_gateways(conn, is_certified=False)
print(gateways)

# Initialize gateway by id
gateway = Gateway(conn, id=GATEWAY_ID)
print(gateway)

# Initialize gateway by name
gateway = Gateway(conn, name=GATEWAY_NAME)
print(gateway)



import pandas as pd
import requests
from packaging import version
from tqdm.auto import tqdm

from mstrio import config
from mstrio.api import reports as reports_api
from mstrio.api.schedules import get_contents_schedule
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.object_management.search_operations import SearchPattern, full_search
from mstrio.users_and_groups.user import User
from mstrio.utils.cache import CacheSource, ContentCacheMixin
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import (
    CertifyMixin,
    CopyMixin,
    DeleteMixin,
    Entity,
    MoveMixin,
    ObjectSubTypes,
    ObjectTypes,
)
from mstrio.utils.filter import Filter
from mstrio.utils.helper import (
    exception_handler,
    fallback_on_timeout,
    get_parallel_number,
    get_valid_project_id,
    response_handler,
    sort_object_properties,
)
from mstrio.utils.parser import Parser
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.translation_mixin import TranslationMixin


def list_reports(
    connection: Connection,
    name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    folder_id: str | None = None,
    **filters,
) -> list[type['Report']] | list[dict]:
    """Get list of Report objects or dicts with them.
    Optionally filter reports by specifying 'name'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of reports being searched
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Report objects
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project_id (string, optional): Project ID
        project_name (string, optional): Project name
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        folder_id (string, optional): ID of a folder where the search
            will be performed. Defaults to None.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']

    Returns:
        list with Report objects or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    objects_ = full_search(
        connection,
        object_types=ObjectTypes.REPORT_DEFINITION,
        project=project_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        root=folder_id,
        **filters,
    )
    reports = [
        item for item in objects_ if Report._is_subtype_supported(item['subtype'])
    ]
    if to_dictionary:
        return reports
    return [
        Report.from_dict(
            source=report_dict, connection=connection, with_missing_value=True
        )
        for report_dict in reports
    ]


class Report(
    Entity,
    CertifyMixin,
    CopyMixin,
    MoveMixin,
    DeleteMixin,
    ContentCacheMixin,
    TranslationMixin,
):
    """Access, filter, publish, and extract data from in-memory reports.

    Create a Report object to load basic information on a report dataset.
    Specify subset of report to be fetched through `Report.apply_filters()` and
    `Report.clear_filters()`. Fetch dataset through `Report.to_dataframe()`
    method.

    _CACHE_TYPE is a variable used by ContentCache class for cache filtering
    purposes.

    Attributes:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id: Identifier of a pre-existing report containing the required
            data.
        name: Report name
        description: Report description
        abbreviation: Report abbreviation
        instance_id: Identifier of an instance if report instance has been
            already initialized, NULL by default.
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: owner User object
        view_media: View media information
        ancestors: List of ancestor folders
        certified_info: Information whether report is certified or not,
            CertifiedInfo object
        attributes: List of attributes
        metrics: List of metrics
        attr_elements: All attributes elements of report
        selected_attributes: IDs of filtered attributes
        selected_metrics: IDs of filtered metrics
        selected_attr_elements: IDs of filtered attribute elements
        dataframe: content of a report extracted into a Pandas `DataFrame`
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.REPORT_DEFINITION
    _OBJECT_SUBTYPES = [
        ObjectSubTypes.REPORT_GRID,
        ObjectSubTypes.REPORT_GRAPH,
        ObjectSubTypes.REPORT_ENGINE,
        ObjectSubTypes.REPORT_GRID_AND_GRAPH,
        ObjectSubTypes.REPORT_TRANSACTION,
        ObjectSubTypes.REPORT_HYPER_CARD,
    ]
    _CACHE_TYPE = CacheSource.Type.REPORT
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'certified_info': CertifiedInfo.from_dict,
    }
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    _API_PATCH: dict = {
        ('name', 'description', 'abbreviation', 'hidden', 'folder_id'): (
            objects_processors.update,
            'partial_put',
        )
    }

    def __init__(
        self,
        connection: Connection,
        id: str,
        instance_id: str | None = None,
        parallel: bool = True,
        progress_bar: bool = True,
    ):
        """Initialize an instance of a report.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing report containing
                the required data.
            instance_id (str): Identifier of an instance if report instance has
                been already initialized, NULL by default.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
            progress_bar(bool, optional): If True (default), show the download
                progress bar.
        """
        if not id:
            raise ValueError("Report ID cannot be an empty string.")

        connection._validate_project_selected()

        super().__init__(
            connection,
            id,
            instance_id=instance_id,
            parallel=parallel,
            progress_bar=progress_bar,
        )

    def _init_variables(self, default_value, **kwargs):
        super()._init_variables(default_value=default_value, **kwargs)
        self.instance_id = kwargs.get("instance_id")
        self._parallel = kwargs.get("parallel", True)
        self._initial_limit = 1000
        self._progress_bar = kwargs.get("progress_bar", True) and config.progress_bar
        self._cross_tab = False
        self._cross_tab_filter = {}
        self._subtotals = {}
        self._dataframe = None
        self._attr_elements = None

        self._attributes = []
        self._metrics = []
        self.__definition_retrieved = False
        self.__filter = None

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        hidden: bool | None = None,
    ):
        """Alter Report properties.

        Args:
            name: new name of the Report
            description: new description of the Report
            abbreviation: new abbreviation of the Report
            hidden: Specifies whether the metric is hidden
        """
        func = self.alter
        args = func.__code__.co_varnames[: func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def to_dataframe(self, limit: int | None = None) -> pd.DataFrame:
        """Extract contents of a report instance into a Pandas `DataFrame`.

        Args:
            limit (None or int, optional): Used to control data extract
                behavior. By default (None) the limit is calculated
                automatically, based on an optimized physical size of one chunk.
                Setting limit manually will force the number of rows per chunk.
                Depending on system resources, a higher limit (e.g. 50,000) may
                reduce the total time required to extract the entire dataset.

        Returns:
            Pandas Data Frame containing the report contents.
        """
        if limit:
            self._initial_limit = limit

        if self.instance_id is None:
            res = self.__initialize_report(self._initial_limit)
        else:
            # try to get first chunk from already initialized instance of report
            # if not possible, initialize new instance
            try:
                res = self.__get_chunk(
                    instance_id=self.instance_id, offset=0, limit=self._initial_limit
                )
            except requests.HTTPError:
                res = self.__initialize_report(self._initial_limit)

        # Gets the pagination totals from the response object
        _instance = res.json()
        self.instance_id = _instance['instanceId']
        paging = _instance['data']['paging']

        # initialize parser and process first response
        p = Parser(response=_instance, parse_cube=False)
        p.parse(response=_instance)

        # If there are more rows to fetch, fetch them
        if paging['current'] != paging['total']:
            if not limit:
                limit = max(
                    1000,
                    int((self._initial_limit * self._SIZE_LIMIT) / len(res.content)),
                )
            # Count the number of additional iterations
            it_total = int((paging['total'] - self._initial_limit) / limit) + (
                (paging['total'] - self._initial_limit) % limit != 0
            )

            if self._parallel and it_total > 1:
                threads = get_parallel_number(it_total)
                with FuturesSessionWithRenewal(
                    connection=self._connection, max_workers=threads
                ) as session:
                    fetch_pbar = tqdm(
                        desc="Downloading",
                        total=it_total + 1,
                        disable=(not self._progress_bar),
                    )
                    future = self.__fetch_chunks_future(
                        session, paging, self.instance_id, limit
                    )
                    fetch_pbar.update()
                    for i, f in enumerate(future, start=1):
                        response = f.result()
                        if not response.ok:
                            response_handler(response, "Error getting report contents.")
                        fetch_pbar.update()
                        fetch_pbar.set_postfix(
                            rows=str(
                                min(self._initial_limit + i * limit, paging['total'])
                            )
                        )
                        p.parse(response.json())
                    fetch_pbar.close()
            else:
                self.__fetch_chunks(p, paging, it_total, self.instance_id, limit)

        # return parsed data as a data frame
        self._dataframe = p.dataframe

        # filter dataframe if report had crosstabs and filters were applied
        if self._cross_tab_filter != {}:
            if self._cross_tab_filter['metrics'] is not None:
                # drop metrics columns from dataframe
                metr_names = [
                    el['name']
                    for el in list(
                        filter(
                            lambda x: x['id'] not in self._cross_tab_filter['metrics'],
                            self.metrics,
                        )
                    )
                ]
                self._dataframe = self._dataframe.drop(metr_names, axis=1)

            if self._cross_tab_filter['attr_elements'] is not None:
                # create dict of attributes and elements to iterate through
                attr_dict = {}
                for attribute in self._cross_tab_filter['attr_elements']:
                    key = attribute[:32]
                    attr_dict.setdefault(key, []).append(attribute[33:])
                # initialize indexes series for filter
                indexes = pd.Series([False] * len(self._dataframe))

                # logical OR for filtered attribute elements
                for attribute in attr_dict:
                    attr_name = list(
                        filter(
                            lambda x, attr=attribute: x['id'] in attr, self.attributes
                        )
                    )[0]['name']
                    elements = attr_dict[attribute]
                    indexes = indexes | self._dataframe[attr_name].isin(elements)
                # select dataframe indexes with
                self._dataframe = self._dataframe[indexes]

            if self._cross_tab_filter['attributes'] is not None:
                attr_names = [
                    el['name']
                    for el in list(
                        filter(
                            lambda x: x['id']
                            not in self._cross_tab_filter['attributes'],
                            self.attributes,
                        )
                    )
                ]
                # filtering out attribute forms columns
                to_be_removed = []
                to_be_added = []
                for attr in attr_names:
                    forms = [
                        column
                        for column in self._dataframe.columns
                        if column.startswith(attr + '@')
                    ]
                    if forms:
                        to_be_removed.append(attr)
                        to_be_added.extend(forms)
                for elem in to_be_removed:
                    attr_names.remove(elem)
                attr_names.extend(to_be_added)
                # drop filtered out columns
                self._dataframe = self._dataframe.drop(attr_names, axis=1)
        return self._dataframe

    def __fetch_chunks_future(self, future_session, pagination, instance_id, limit):
        """Fetch added rows from this object instance from the Intelligence
        Server."""
        return [
            reports_api.report_instance_id_coroutine(
                future_session,
                connection=self._connection,
                report_id=self._id,
                instance_id=instance_id,
                offset=_offset,
                limit=limit,
            )
            for _offset in range(self._initial_limit, pagination['total'], limit)
        ]

    def __fetch_chunks(self, parser, pagination, it_total, instance_id, limit):
        """Fetch added rows from this object instance from the Intelligence
        Server."""
        with tqdm(
            desc="Downloading", total=it_total + 1, disable=(not self._progress_bar)
        ) as fetch_pbar:
            fetch_pbar.update()
            for _offset in range(self._initial_limit, pagination['total'], limit):
                response = self.__get_chunk(
                    instance_id=instance_id, offset=_offset, limit=limit
                )
                fetch_pbar.update()
                fetch_pbar.set_postfix(
                    rows=str(min(_offset + limit, pagination['total']))
                )
                parser.parse(response=response.json())

    def __initialize_report(self, limit: int) -> requests.Response:
        inst_pbar = tqdm(
            desc='Initializing an instance of a report. Please wait...',
            bar_format='{desc}',
            leave=False,
            ncols=285,
            disable=(not self._progress_bar),
        )

        # Switch off subtotals if I-Server version is higher than 11.2.1
        body = self._filter._request_body()
        if version.parse(self._connection.iserver_version) >= version.parse(
            "11.2.0100"
        ):
            self._subtotals["visible"] = False
            body["subtotals"] = {"visible": self._subtotals["visible"]}

        # Request a new instance, set instance id
        response = reports_api.report_instance(
            connection=self._connection,
            report_id=self._id,
            body=body,
            offset=0,
            limit=limit or self._initial_limit,
        )
        inst_pbar.close()
        return response

    def __get_chunk(
        self, instance_id: str, offset: int, limit: int
    ) -> requests.Response:
        return reports_api.report_instance_id(
            connection=self._connection,
            report_id=self._id,
            instance_id=instance_id,
            offset=offset,
            limit=limit,
        )

    def apply_filters(
        self,
        attributes: list | None = None,
        metrics: list | None = None,
        attr_elements: list | None = None,
        operator: str = 'In',
    ) -> None:
        """Apply filters on the reports's objects.

        Filter by attributes, metrics and attribute elements.

        Args:
            attributes (list or None, optional): ids of attributes to be
                included in the filter. If list is empty, no attributes will be
                selected and metric data will be aggregated.
            metrics (list or None, optional): ids of metrics to be included in
                the filter. If list is empty, no metrics will be selected.
            attr_elements (list or None, optional): attribute elements to be
                included in the filter.
            operator (str, optional): a str flag used to specify if the
                attribute elements selected inside the filter should be included
                or excluded. Allowed values are: 'In', 'NotIn'.
        """
        filtering_is_requested = bool(
            not all(element is None for element in [attributes, metrics, attr_elements])
        )

        if self._cross_tab:
            self._cross_tab_filter = {
                'attributes': attributes,
                'metrics': metrics,
                'attr_elements': attr_elements,
            }
        elif filtering_is_requested:
            self._filter._clear(
                attributes=attributes, metrics=metrics, attr_elements=attr_elements
            )
            self._filter.operator = operator
            self._select_attribute_filter_conditionally(attributes)
            self._select_metric_filter_conditionally(metrics)
            self._select_attr_el_filter_conditionally(attr_elements)
            # Clear instance, to generate new with new filters
            self.instance_id = None

    def _select_attribute_filter_conditionally(self, attributes_filtered) -> None:
        if attributes_filtered:
            self._filter._select(object_id=attributes_filtered)
        elif attributes_filtered is not None:
            self._filter.attr_selected = []

    def _select_metric_filter_conditionally(self, metrics_filtered) -> None:
        if metrics_filtered:
            self._filter._select(object_id=metrics_filtered)
        elif metrics_filtered is not None:
            self._filter.metr_selected = []

    def _select_attr_el_filter_conditionally(self, attr_el_filtered) -> None:
        if attr_el_filtered is not None:
            self._filter._select_attr_el(element_id=attr_el_filtered)

    def clear_filters(self) -> None:
        """Clear previously set filters, allowing all attributes, metrics, and
        attribute elements to be retrieved."""

        self._filter._clear()
        if self._cross_tab:
            self._filter._select(object_id=[el['id'] for el in self.attributes])
            self._filter._select(object_id=[el['id'] for el in self.metrics])
        # Clear instance, to generate new with new filters
        self.instance_id = None

    def _get_definition(self) -> None:
        """Get the definition of a report, including attributes and metrics.

        Implements GET /v2/reports/<report_id>.
        """
        response = reports_api.report_definition(
            connection=self._connection, report_id=self._id
        ).json()

        grid = response["definition"]["grid"]
        available_objects = response['definition']['availableObjects']

        if version.parse(self._connection.iserver_version) >= version.parse(
            "11.2.0100"
        ):
            self._subtotals = grid["subtotals"]
        self.name = response["name"]
        self._cross_tab = grid["crossTab"]

        # Check if report have custom groups or consolidations
        if available_objects['customGroups']:
            exception_handler(
                msg="Reports with custom groups are not supported.",
                exception_type=ImportError,
            )
        if available_objects['consolidations']:
            exception_handler(
                msg="Reports with consolidations are not supported.",
                exception_type=ImportError,
            )

        full_attributes = []
        for row in grid["rows"]:
            if row["type"] == "attribute":
                full_attributes.append(row)
        for column in grid["columns"]:
            if column["type"] == "attribute":
                full_attributes.append(column)
        self._attributes = [
            {'name': attr['name'], 'id': attr['id']} for attr in full_attributes
        ]

        # Retrieve metrics from the report grid (only selected metrics)
        metrics_position = grid.get("metricsPosition")
        if metrics_position is None:
            self._metrics = []
        else:
            full_metrics = grid[metrics_position["axis"]][metrics_position["index"]][
                "elements"
            ]
            self._metrics = [
                {'name': metr['name'], 'id': metr['id']} for metr in full_metrics
            ]

        self.__definition_retrieved = True

    def __get_attr_elements(self, limit: int = 50000) -> list:
        """Get elements of report attributes synchronously.

        Implements GET /reports/<report_id>/attributes/<attribute_id>/elements.
        """

        def fetch_for_attribute(attribute):
            @fallback_on_timeout()
            def fetch_for_attribute_given_limit(limit):
                response = reports_api.report_single_attribute_elements(
                    connection=self._connection,
                    report_id=self._id,
                    attribute_id=attribute['id'],
                    offset=0,
                    limit=limit,
                )
                # Get total number of rows from headers.
                total = int(response.headers['x-mstr-total-count'])
                # Get attribute elements from the response.
                elements = response.json()

                # If total number of elements is bigger than the chunk size
                # (limit), fetch them incrementally.
                for _offset in range(limit, total, limit):
                    response = reports_api.report_single_attribute_elements(
                        connection=self._connection,
                        report_id=self._id,
                        attribute_id=attribute['id'],
                        offset=_offset,
                        limit=limit,
                    )
                    elements.extend(response.json())

                # Return attribute data.
                return {
                    "attribute_name": attribute['name'],
                    "attribute_id": attribute['id'],
                    "elements": elements,
                }

            return fetch_for_attribute_given_limit(limit)[0]

        attr_elements = []
        if self.attributes:
            pbar = tqdm(
                self.attributes,
                desc="Loading attribute elements",
                leave=False,
                disable=(not self._progress_bar),
            )
            attr_elements = [fetch_for_attribute(attribute) for attribute in pbar]
            pbar.close()

        return attr_elements

    def __get_attr_elements_async(self, limit: int = 50000) -> list:
        """Get elements of report attributes asynchronously.

        Implements GET /reports/<report_id>/attributes/<attribute_id>/elements.
        """

        attr_elements = []
        if self.attributes:
            threads = get_parallel_number(len(self.attributes))
            with FuturesSessionWithRenewal(
                connection=self._connection, max_workers=threads
            ) as session:
                # Fetch first chunk of attribute elements.
                futures = self.__fetch_attribute_elements_chunks(session, limit)
                pbar = tqdm(
                    futures,
                    desc="Loading attribute elements",
                    leave=False,
                    disable=(not self._progress_bar),
                )
                for i, future in enumerate(pbar):
                    attr = self.attributes[i]
                    response = future.result()
                    if not response.ok:
                        response_handler(
                            response, f"Error getting attribute {attr['name']} elements"
                        )
                    elements = response.json()
                    # Get total number of rows from headers.
                    total = int(response.headers['x-mstr-total-count'])
                    for _offset in range(limit, total, limit):
                        response = reports_api.report_single_attribute_elements(
                            connection=self._connection,
                            report_id=self._id,
                            attribute_id=attr["id"],
                            offset=_offset,
                            limit=limit,
                        )
                        elements.extend(response.json())
                    # Append attribute data to the list of attributes.
                    attr_elements.append(
                        {
                            "attribute_name": attr['name'],
                            "attribute_id": attr['id'],
                            "elements": elements,
                        }
                    )
                pbar.close()

            return attr_elements

    def __fetch_attribute_elements_chunks(self, future_session, limit: int) -> list:
        # Fetch add'l rows from this object instance
        return [
            reports_api.report_single_attribute_elements_coroutine(
                future_session,
                connection=self._connection,
                report_id=self._id,
                attribute_id=attribute['id'],
                offset=0,
                limit=limit,
            )
            for attribute in self.attributes
        ]

    def list_properties(self):
        """List all properties of the object."""

        attributes = {
            key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')
        }
        attributes = {
            **attributes,
            "id": self.id,
            "instance_id": self.instance_id,
            "type": self.type,
            "subtype": self.subtype,
            "ext_type": self.ext_type,
            "date_created": self.date_created,
            "date_modified": self.date_modified,
            "version": self.version,
            "owner": self.owner,
            "view_media": self.view_media,
            "ancestors": self.ancestors,
            "certified_info": self.certified_info,
            "acg": self.acg,
            "acl": self.acl,
            "attributes": self.attributes,
            "metrics": self.metrics,
        }
        return {
            key: attributes[key]
            for key in sorted(attributes, key=sort_object_properties)
        }

    def list_available_schedules(
        self, to_dictionary: bool = False
    ) -> list["Schedule"] | list[dict]:
        """Get a list of schedules available for the report.

        Args:
            to_dictionary (bool, optional): If True returns a list of
                dictionaries, otherwise returns a list of Schedules.
                False by default.

        Returns:
            List of Schedule objects or list of dictionaries.
        """
        schedules_list_response = (
            get_contents_schedule(
                connection=self.connection,
                project_id=self.connection.project_id,
                body={'id': self.id, 'type': 'report'},
            )
            .json()
            .get('schedules')
        )
        if to_dictionary:
            return schedules_list_response
        return [
            Schedule.from_dict(connection=self.connection, source=schedule_id)
            for schedule_id in schedules_list_response
        ]

    @property
    def attributes(self):
        if not self.__definition_retrieved:
            self._get_definition()
        return self._attributes

    @property
    def metrics(self):
        if not self.__definition_retrieved:
            self._get_definition()
        return self._metrics

    @property
    def attr_elements(self):
        if not self.__definition_retrieved:
            self._get_definition()
        if not self._attr_elements and self._id:
            if self._parallel is True:
                # TODO: move the fallback inside the function to apply
                # per-attribute, like with non-async version.
                self._attr_elements = fallback_on_timeout()(
                    self.__get_attr_elements_async
                )(50000)[0]
            else:
                self._attr_elements = self.__get_attr_elements()
            self._filter._populate_attr_elements(self._attr_elements)
        return self._attr_elements

    @property
    def _filter(self):
        if not self.__definition_retrieved:
            self._get_definition()
        if self.__filter is None:
            self.__filter = Filter(
                attributes=self._attributes,
                metrics=self._metrics,
                attr_elements=self._attr_elements,
            )
        return self.__filter

    @property
    def selected_attributes(self):
        """Selected attributes for filtering."""
        return self._filter.attr_selected

    @property
    def selected_metrics(self):
        """Selected metrics for filtering."""
        return self._filter.metr_selected

    @property
    def selected_attr_elements(self):
        """Selected attribute elements for filtering."""
        return self._filter.attr_elem_selected

    @property
    def dataframe(self) -> pd.DataFrame:
        if self._dataframe is None:
            exception_handler(
                msg="Dataframe not loaded. Retrieve with Report.to_dataframe().",
                exception_type=Warning,
            )
        return self._dataframe



"""This is the demo script to show how to manage transformations, transformation
attributes and transformation attribute forms. Its basic goal is to present what
can be done with this module and to ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.modeling.expression.enums import ExpressionFormat
from mstrio.modeling.expression.expression import Expression, Token
from mstrio.modeling.schema.helpers import ObjectSubType, SchemaObjectReference
from mstrio.modeling.schema.schema_management import SchemaManagement, SchemaUpdateType
from mstrio.modeling.schema.transformation.transformation import (
    list_transformations, Transformation, TransformationAttribute, TransformationAttributeForm
)

# For every object we want to reference using a SchemaObjectReference we need
# to provide an Object ID for. For the script to work correctly all occurences
# of `'<object_id>'` and others with form `<some_name>` need to be replaced with
# data specific to the object used.

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Define variables which can be later used in a script
FOLDER_ID = $folder_id  # Insert folder ID here
TRANSFORMATION_NAME = $transformation_name  # Insert name of edited transformation here

# Example of Transformation data
# Parts of this dictionary will be used in the later parts of this demo script
TRANSFORMATION_DATA = {
    'name': TRANSFORMATION_NAME,
    'destination_folder': FOLDER_ID,
    'sub_type': ObjectSubType.ROLE_TRANSFORMATION,
    'attributes': [
        TransformationAttribute(
            id='<object_id>',
            base_attribute=SchemaObjectReference(
                object_id='<object_id>',
                sub_type=ObjectSubType.ATTRIBUTE,
                name="<attribute_name>"
            ),
            forms=[
                TransformationAttributeForm(
                    id='<object_id>',
                    name='<transformation_attribute_form_name>',
                    lookup_table=SchemaObjectReference(
                        object_id='<object_id>',
                        sub_type=ObjectSubType.LOGICAL_TABLE,
                        name='<logical_table_name>'
                    ),
                    expression=Expression(
                        tokens=[
                            Token(value='<token_value>', type='column_reference'),
                            Token(value='', type='end_of_text')
                        ]
                    )
                )
            ]
        )
    ],
    'mapping_type': 'one_to_one'
}

# Example TransformationAttributeForm data
# This object will be used in the later parts of this demo script
TRANSFORMATION_ATTRIBUTE_FORM = TransformationAttributeForm(
    id='<object_id>',
    name='<transformation_attribute_form_name>',
    lookup_table=SchemaObjectReference(
        object_id='<object_id>',
        sub_type=ObjectSubType.LOGICAL_TABLE,
        name='<logical_table_name>'
    ),
    expression=Expression(
        text='<text>',
        tokens=[
            Token(value='<token_value>', type='column_reference'),
            Token(value='', type='end_of_text')
        ]
    )
)

# Example TransformationAttribute data
# This object will nbe used in the later parts of this demo script
TRANSFORMATION_ATTRIBUTE = TransformationAttribute(
    id='<object_id>',
    base_attribute=SchemaObjectReference(
        object_id='<object_id>',
        sub_type=ObjectSubType.ATTRIBUTE,
        name='<attribute_name>'
    ),
    forms=[
        TransformationAttributeForm(
            id='<object_id>',
            name='<transformation_attribute_form_name>',
            lookup_table=SchemaObjectReference(
                object_id='<object_id>',
                sub_type=ObjectSubType.LOGICAL_TABLE,
                name='<logical_table_name>'
            ),
            expression=Expression(
                text='<expression_text>',
                tokens=[
                    Token(value='<token_value>', type='column_reference'),
                    Token(value='', type='end_of_text')
                ]
            )
        )
    ]
)

# Transformation management
# Get list of transformations, with examples of different conditions
# with expressions represented as tree
list_of_all_transformations = list_transformations(connection=conn)
list_of_limited_transformations = list_transformations(connection=conn, limit=5)
list_of_transformations_by_name = list_transformations(connection=conn, name=TRANSFORMATION_NAME)

# Define a variable which can be later used in a script
TRANSFORMATION_ID = $transformation_id  # Insert ID of transformation here

list_of_transformations_by_id = list_transformations(connection=conn, id=TRANSFORMATION_ID)
list_of_transformations_as_trees = list_transformations(
    connection=conn, show_expression_as=ExpressionFormat.TREE
)

# List Transformations with expressions as tokens
list_of_transformations_as_tokens = list_transformations(
    connection=conn, show_expression_as=ExpressionFormat.TOKENS
)

# Get specific Transformation by id with expression represented as trees (default)
transf = Transformation(connection=conn, id=TRANSFORMATION_ID, show_expression_as=ExpressionFormat.TREE)
# Get specific Transformation by id with expression represented as tokens
transf = Transformation(connection=conn, id=TRANSFORMATION_ID, show_expression_as=ExpressionFormat.TOKENS)
# Get specific Transformation by name with expression represented as trees (default)
transf = Transformation(connection=conn, name=TRANSFORMATION_NAME, show_expression_as=ExpressionFormat.TREE)
# Get specific Transformation by name with expression represented as tokens
transf = Transformation(connection=conn, name=TRANSFORMATION_NAME, show_expression_as=ExpressionFormat.TOKENS)

# Listing properties
properties = transf.list_properties()

# Create transformation and get it with expression represented as tree
transf = Transformation.create(connection=conn, **TRANSFORMATION_DATA)
transf = Transformation.create(
    connection=conn,
    sub_type=TRANSFORMATION_DATA['sub_type'],
    name=TRANSFORMATION_DATA['name'],
    destination_folder=TRANSFORMATION_DATA['destination_folder'],
    attributes=TRANSFORMATION_DATA['attributes'],
    mapping_type=TRANSFORMATION_DATA['mapping_type']
)
transf = Transformation.create(
    connection=conn, **TRANSFORMATION_DATA, show_expression_as=ExpressionFormat.TREE
)

# Create transformation and get it with expression represented as tokens
transf = Transformation.create(
    connection=conn, **TRANSFORMATION_DATA, show_expression_as=ExpressionFormat.TOKENS
)

# Define variables which can be later used in a script
TRANSFORMATION_NEW_NAME = $transformation_new_name  # Insert new name of edited transformation here
TRANSFORMATION_NEW_DESCRIPTION = $transformation_new_description  # Insert new description of edited transformation here

# Alter transformations
transf.alter(name=TRANSFORMATION_NEW_NAME)
transf.alter(description=TRANSFORMATION_NEW_DESCRIPTION)

# Alter TransformationAttribute
transf.alter(attributes=[TRANSFORMATION_ATTRIBUTE])

# Alter TransformationAttributeForm
# Each TransformationAttribute can store only one TransformationAttributeForm
attr = transf.attributes[0]
attr.forms = [TRANSFORMATION_ATTRIBUTE_FORM]
transf.alter(attributes=[attr])

# Deleting Transformation
transf.delete(force=True)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dashboards and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])



"""This is the demo script to show how to import data from OlapCubes and
Reports. It is possible to select attributes, metrics and attribute forms which
is presented.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.project_objects import OlapCube, Report
import pandas as pd
import datetime as dt

# get connection to an environment
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
CUBE_ID = $cube_id
REPORT_ID = $report_id

# get cube based on its id and store it in data frame
my_cube = OlapCube(connection=conn, id=CUBE_ID)
my_cube_df = my_cube.to_dataframe()

# if any of your columns should have type 'Time' please set it manually
# Column name should be set to the name of column that should have type 'Time'.
# Please repeat this line for all of the columns that should have type 'Time'.
COLUMN_NAME = $column_name
my_cube_df[COLUMN_NAME] = pd.to_datetime(my_cube_df[COLUMN_NAME], format='%H:%M:%S').dt.time
# If the values in column name are in HH:MM format and not in HH:MM:SS
# you can format it as below
my_cube_df[COLUMN_NAME] = pd.to_datetime(my_cube_df[COLUMN_NAME], format='%H:%M').dt.time

# get report based on its id and store it in data frame
my_report = Report(connection=conn, id=REPORT_ID, parallel=False)
my_report_df = my_report.to_dataframe()

# get list of ids of metrics, attributes or attribute elements available within
# Cube or Report
my_cube.metrics
my_cube.attributes
my_cube.attr_elements

# Define variables which can be later used in a script
ATTRIBUTES_LIST = $attributes_list  # Insert list of attributes that you want to include in  your functions
METRICS_LIST = $metrics_list  # insert list of metrics to be used in your functions
ATTRIBUTES_ELEMENTS_LIST = $attributes_elements_list  # insert list of attributes to be used in your functions

# by default all elements are shown in the data frame. To choose elements you
# have to pass proper IDs to function 'apply_filters()' which is available for
# Cube and Report
my_cube.apply_filters(
    attributes=ATTRIBUTES_LIST,
    metrics=METRICS_LIST,
    attr_elements=ATTRIBUTES_ELEMENTS_LIST,
)
# check selected elements which will be placed into a dataframe
my_cube.selected_attributes
my_cube.selected_metrics
my_cube.selected_attr_elements

my_cube_applied_filters_df = my_cube.to_dataframe()

# to exclude specific attribute elements, pass the `operator="NotIn"` to
# `apply_filters()` method
my_cube.apply_filters(
    attributes=ATTRIBUTES_LIST,
    metrics=METRICS_LIST,
    attr_elements=ATTRIBUTES_ELEMENTS_LIST,
    operator="NotIn",
)

my_cube_excluded_elements_df = my_cube.to_dataframe()



from mstrio.api import languages as languages_api
from mstrio.connection import Connection
from mstrio.utils.helper import fetch_objects


def get(connection: Connection, id: str) -> dict:
    """Get language by a specified ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the language

    Returns:
        dict representing language object
    """
    return languages_api.get_language(connection=connection, id=id).json()


def get_all(connection: Connection, limit: int, filters) -> list[dict]:
    """Get list of languages.

    Args:
        connection: MicroStrategy REST API connection object
        limit: limit of languages to list
        filters: filters

    Returns:
        list of dicts representing languages"""
    return fetch_objects(
        connection=connection,
        api=languages_api.list_languages,
        limit=limit,
        filters=filters,
        dict_unpack_value="languages",
    )


def get_interface_languages(connection: Connection) -> list[dict]:
    """Get list of interface languages.

    Args:
        connection: MicroStrategy REST API connection object

    Returns:
        list of dicts representing interface languages"""
    return (
        languages_api.list_interface_languages(connection=connection)
        .json()
        .get('interfaceLanguages')
    )


def get_formatting_settings(connection: Connection, id: str) -> dict:
    """Get formatting settings for a language with specified ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the language

    Returns:
        dict representing language object formatting settings
    """
    return (
        languages_api.get_language_formatting_settings(connection=connection, id=id)
        .json()
        .get('formattingSettings')
        .get('timeInterval')
    )


def update_formatting_settings(connection: Connection, id: str, body: dict) -> dict:
    """Update formatting settings for a language with specified ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the language
        body: json body for the request

    Returns:
        dict representing language object formatting settings
    """
    return (
        languages_api.update_language_formatting_settings(
            connection=connection, id=id, body=body
        )
        .json()
        .get('formattingSettings')
        .get('timeInterval')
    )


def create(connection: Connection, body: dict) -> dict:
    """Create a language based on provided body.

    Args:
        connection: MicroStrategy REST API connection object
        body: dictionary with language details

    Returns:
        dict representing language object"""
    return languages_api.create_language(connection=connection, body=body).json()


def update(connection: Connection, id: str, body: dict) -> dict:
    """Update a language with specified ID.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the language
        body: json body for the request

    Returns:
        dict representing language object
    """
    return languages_api.update_language(connection=connection, id=id, body=body).json()



from dataclasses import dataclass

from mstrio.utils.helper import Dictable


@dataclass
class Node(Dictable):
    name: str | None = None
    address: str | None = None
    service_control: bool | None = None

    @classmethod
    def from_dict(cls, source: dict):
        if not source.get('name'):
            source['name'] = source['node']
        return super().from_dict(source)


@dataclass
class Service(Dictable):
    id: str
    service: str
    port: int
    status: str
    tags: dict | None = None
    output: str | None = None


@dataclass
class ServiceWithNode(Dictable):
    node: str
    address: str
    id: str
    service_control: bool
    port: int
    status: str
    service_address: str
    tags: dict | None = None
    output: str | None = None



from mstrio.connection import get_connection
from mstrio.users_and_groups.contact import list_contacts
from mstrio.users_and_groups.user import User

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# migrate addresses from one User to another
user_1 = User(conn, name='User Korean')
user_2 = User(conn, name='Tilda Austin')
for address in user_1.addresses:
    print(address)
    user_2.add_address(address=address['id'])
    user_1.remove_address(id=address['id'])

# migrate User contacts to another User
list_contacts = list_contacts(conn, linked_user=user_1)
for contact in list_contacts:
    contact.alter(linked_user=user_2)



import logging
from typing import Optional, Union

from pandas import DataFrame

from mstrio import config
from mstrio.api import monitors
from mstrio.server.project import Project, compare_project_settings
from mstrio.server.server import ServerSettings
from mstrio.utils import helper
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@class_version_handler('11.3.0000')
class Environment:
    """Browse and manage Projects on the environment. List loaded
    projects, nodes (servers) and compare project settings on the
    environment. Browse and modify I-Server settings.

    Attributes:
        connection: A MicroStrategy connection object.
        server_settings: Intelligence Server settings object.
        nodes: List of I-Server nodes and their properties.
        node_names: List of I-Server node names.
    """

    def __init__(self, connection):
        """Initialize Environment object.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
        """
        self.connection = connection
        self._nodes = None

    @property
    def server_settings(self) -> ServerSettings:
        """`ServerSettings` object storing I-Server settings.

        Settings can be listed by using `list_properties()` method.
        Settings can be modified directly by settings the values in the
        object.
        """
        if not hasattr(self, "_server_settings"):
            self._server_settings = ServerSettings(self.connection)
        return self._server_settings

    @server_settings.setter
    def server_settings(self, server_settings: ServerSettings) -> None:
        self._server_settings = server_settings

    def update_settings(self):
        """Update the current server settings saved in the ServerSettings
        object."""
        self.server_settings.update()

    def fetch_settings(self) -> None:
        """Fetch the current server settings from the environment."""
        self.server_settings.fetch()

    def create_project(
        self, name: str, description: str | None = None, force: bool = False
    ) -> Optional["Project"]:
        """Create a new project on the environment.

        Args:
            name: Name of Project.
            description: Description of Application.
            force: If `True`, overrides the prompt.
        """
        return Project._create(self.connection, name, description, force)

    def list_projects(
        self, to_dictionary: bool = False, limit: int | None = None, **filters
    ) -> list["Project"] | list[dict]:
        """Return list of project objects or project dicts if
        `to_dictionary=True`. Optionally filter the Projects by specifying
        the `filters` keyword arguments.

        Args:
            to_dictionary: If True returns list of project dicts.
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['name', 'id',
                'description', 'date_created', 'date_modified', 'owner'].
        """
        return Project._list_projects(
            connection=self.connection,
            to_dictionary=to_dictionary,
            limit=limit,
            **filters,
        )

    def list_loaded_projects(
        self, to_dictionary: bool = False, **filters
    ) -> list["Project"] | list[dict]:
        """Return list of all loaded project objects or project dicts
        if `to_dictionary=True` that the user has access to. Optionally filter
        the Projects by specifying the `filters` keyword arguments.

        Args:
            to_dictionary: If True, returns list of project dicts
            **filters: Available filter parameters: ['acg', 'id', 'name',
                'status', 'alias', 'description', 'date_created',
                'date_modified', 'owner']
        """
        return Project._list_loaded_projects(
            connection=self.connection,
            to_dictionary=to_dictionary,
            **filters,
        )

    @method_version_handler('11.3.0800')
    def delete_server_object_cache(self) -> None:
        """Delete object cache for all projects on the environment."""
        for project in self.list_loaded_projects():
            project.delete_object_cache()
        if config.verbose:
            logger.info('Server object cache deleted.')

    @method_version_handler('11.3.0800')
    def delete_server_element_cache(self) -> None:
        """Delete element cache for all projects on the environment."""
        for project in self.list_loaded_projects():
            project.delete_element_cache()
        if config.verbose:
            logger.info('Server element cache deleted.')

    def list_nodes(
        self,
        project: Union[str, "Project"] | None = None,
        node_name: str | None = None,
    ) -> list[dict]:
        """Return a list of I-Server nodes and their properties. Optionally
        filter by `project` or `node_name`.

        Args:
            project: ID of project or Project object
            node_name: Name of node
        """
        project_id = project.id if isinstance(project, Project) else project
        response = monitors.get_node_info(self.connection, project_id, node_name).json()
        return response['nodes']

    def is_loaded(
        self, project_id: str | None = None, project_name: str | None = None
    ) -> bool:
        """Check if project is loaded, by passing project ID or name,
        returns True or False.

        Args:
            project_id: Project ID
            project_name: Project name
        """
        if project_id is None and project_name is None:
            helper.exception_handler(
                "Please specify either 'project_name' or 'project_id' argument."
            )
        if project_id is None:
            project_list = Project._list_project_ids(self.connection, name=project_name)
            if project_list:
                project_id = project_list[0]
            else:
                msg = f"There is no project with the given name: '{project_name}'"
                raise ValueError(msg)

        nodes = self.list_nodes(project=project_id)
        loaded = False
        for node in nodes:
            status = node['projects'][0]['status']
            loaded = status == 'loaded'
            if loaded:
                break
        return loaded

    def compare_settings(
        self,
        projects: list[str] | list["Project"] = None,
        show_diff_only: bool = False,
    ) -> DataFrame:
        """Compare project' settings to the first project in the
        provided list.

        Args:
            projects (list of names or project objects, optional): List
                of project objects or names to be compared. First element of
                list is the one to which the rest is compared. If None, all
                projects on the environment will be compared.
            show_diff_only(bool, optional): Whether to display all settings or
                only different from first project in list.

        Returns:
            Dataframe with values of selected project' settings.
        """

        def not_exist_warning(wrong_name):
            helper.exception_handler(
                f"Project '{wrong_name}' does not exist and will be skipped.",
                exception_type=Warning,
            )

        if projects:
            just_objects = [
                project for project in projects if isinstance(project, Project)
            ]
        else:
            just_objects = []
        if len(just_objects) == len(projects):
            return compare_project_settings(projects, show_diff_only)

        all_projects = self.list_projects()
        if isinstance(projects, list):
            if len(projects) < 2:
                helper.exception_handler(
                    "Provide more than one project object or name in list",
                    exception_type=TypeError,
                )

            # extract project names from either project object or strings
            project_names = [
                project.name if isinstance(project, Project) else str(project)
                for project in projects
            ]
            # filter only valid projects
            all_project_names = [project.name for project in all_projects]
            [
                not_exist_warning(a_name)
                for a_name in project_names
                if a_name not in all_project_names
            ]

            projects = list(
                filter(lambda project: project.name in project_names, all_projects)
            )
            projects = sorted(projects, key=lambda x: project_names.index(x.name))

        elif projects is None:
            projects = all_projects
        else:
            helper.exception_handler(
                "The 'projects' parameter needs to be a list of len > 1 or None.",
                exception_type=TypeError,
            )

        return compare_project_settings(projects, show_diff_only)

    def is_cluster(self):
        return len(self.nodes) > 1

    @property
    def nodes(self):
        if not self._nodes:
            self._nodes = self.list_nodes()

        return self._nodes

    @property
    def node_names(self):
        return [node['name'] for node in self.nodes]



from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg="Error getting transformation with ID: {id}")
def get_transformation(
    connection: Connection,
    id: str,
    changeset_id: str | None = None,
    show_expression_as: list[str] | None = None,
):
    """Get definition of a single transformation by id

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of an transformation
        changeset_id: ID of a changeset
        show_expression_as: Specifies the format in which the expressions
           are returned in response
           Available values: 'tokens', 'tree'
           If omitted, the expression is returned in 'text' format
           If 'tree', the expression is returned in 'text' and 'tree' formats.
           If 'tokens', the expression is returned in 'text' and 'tokens'
           formats.

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.get(
        endpoint=f'/api/model/transformations/{id}',
        headers={'X-MSTR-MS-Changeset': changeset_id},
        params={'showExpressionAs': show_expression_as},
    )


@unpack_information
@ErrorHandler(err_msg="Error creating a transformation")
def create_transformation(
    connection: Connection, body: dict, show_expression_as: list[str] | None = None
):
    """Create a new transformation in the changeset,
    based on the definition provided in request body.

    Args:
        connection: MicroStrategy REST API connection object
        body: Transformation creation data
        show_expression_as: Specifies the format in which the expressions are
           returned in response
           Available values: 'tokens', 'tree'
           If omitted, the expression is returned in 'text' format
           If 'tree', the expression is returned in 'text' and 'tree' formats.
           If 'tokens', the expression is returned in 'text' and 'tokens'
           formats.

    Return:
        HTTP response object. Expected status: 200
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/transformations',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={'showExpressionAs': show_expression_as},
            json=body,
        )


@unpack_information
@ErrorHandler(err_msg="Error updating transformation with ID: {id}")
def update_transformation(
    connection: Connection,
    id: str,
    body: dict,
    show_expression_as: list[str] | None = None,
):
    """Update a specific transformation in the changeset,
    based on the definition provided in the request body.
    It returns the transformation's updated definition in the changeset.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a Transformation
        body: Transformation update data
        show_expression_as: Specifies the format in which the expressions
           are returned in response.
           Available values: 'tokens', 'tree'
           If omitted, the expression is returned in 'text' format
           If 'tree', the expression is returned in 'text' and 'tree' formats.
           If 'tokens', the expression is returned in 'text' and 'tokens'
           formats.

    Return:
        HTTP response object. Expected status: 200
    """
    with changeset_manager(connection) as changeset_id:
        return connection.patch(
            endpoint=f'/api/model/transformations/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={'showExpressionAs': show_expression_as},
            json=body,
        )



"""This is the demo script meant to show how administrator can perform a
migration of objects from one environment to another.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.

`mstrio.object_management.migration` module is available as a Functionality Preview.
It is subject to change until it is released as Generally Available.
"""

from mstrio.access_and_security.privilege import Privilege
from mstrio.connection import Connection
from mstrio.object_management.migration import (
    bulk_full_migration,
    bulk_migrate_package,
    Migration,
    PackageConfig,
    PackageContentInfo,
    PackageSettings
)
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User

# Define variables which can be later used in a script
PROJECT_NAME = $project_name
SOURCE_BASE_URL = $source_base_url  # usually ends with /MicroStrategyLibrary/api
SOURCE_USERNAME = $source_username
SOURCE_PASSWORD = $source_password
TARGET_BASE_URL = $target_base_url  # usually ends with /MicroStrategyLibrary/api
TARGET_USERNAME = $target_username
TARGET_PASSWORD = $target_password

# Create connections to both source and target environments
source_conn = Connection(
    SOURCE_BASE_URL, SOURCE_USERNAME, SOURCE_PASSWORD, project_name=PROJECT_NAME, login_mode=1
)
target_conn = Connection(
    TARGET_BASE_URL, TARGET_USERNAME, TARGET_PASSWORD, project_name=PROJECT_NAME, login_mode=1
)

# Define a variable which can be later used iin a script
USERNAME = $username  # user that is executing the migration

# Make sure the current user have the following privileges:
#   'Create package', id: 295
#   'Manage Migration Packages'(IServer >=11.3.7) or 'Apply package' (IServer <11.3.7),  id: 296
# They can be granted by admin with the following commands:
user = User(source_conn, username=USERNAME)
Privilege(source_conn, id=295).add_to_user(user)
Privilege(source_conn, id=296).add_to_user(user)

# Or by name:
user2 = User(target_conn, username=USERNAME)
Privilege(target_conn, name='Create package').add_to_user(user2)
Privilege(target_conn, name='Manage Migration Packages').add_to_user(user2)

# Define variables which can be later used in a script
DASHBOARD_ID = $dashboard_id
REPORT_ID = $report_id

# Create PackageConfig with information what object should be migrated and how.
# The options are of type Enum with all possible values listed.
package_settings = PackageSettings(
    PackageSettings.DefaultAction.USE_EXISTING,
    PackageSettings.UpdateSchema.RECAL_TABLE_LOGICAL_SIZE,
    PackageSettings.AclOnReplacingObjects.REPLACE,
    PackageSettings.AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
)
package_content_info = PackageContentInfo(
    id=REPORT_ID,
    type=ObjectTypes.REPORT_DEFINITION,
    action=PackageContentInfo.Action.USE_EXISTING,
    include_dependents=True,
)
package_content_info2 = PackageContentInfo(
    id=DASHBOARD_ID,
    type=ObjectTypes.DOCUMENT_DEFINITION,
    action=PackageContentInfo.Action.USE_EXISTING,
    include_dependents=True,
)

package_config = PackageConfig(
    PackageConfig.PackageUpdateType.PROJECT, package_settings, package_content_info
)
package_config2 = PackageConfig(
    PackageConfig.PackageUpdateType.PROJECT, package_settings, package_content_info2
)

# Define variables which can be later used in a script
SAVE_PATH = $save_path
CUSTOM_PACKAGE_PATH = $custom_package_path

# Create Migrations objects that can use all the functionalities
mig = Migration(
    save_path=SAVE_PATH,
    source_connection=source_conn,
    target_connection=target_conn,
    configuration=package_config,
)

# Create Migration object that can only use `create_package()`
mig2 = Migration(
    save_path=SAVE_PATH,
    source_connection=source_conn,
    configuration=package_config2,
)

# Create Migration object that can only be used for `migrate_package()`
mig3 = Migration(
    target_connection=target_conn,
    custom_package_path=CUSTOM_PACKAGE_PATH,
)

# Short version
# Create import package and save it to the file
mig.create_package()
# or
mig2.create_package()

# Migrate downloaded package to the target environment.
# Create undo package and save it to file
mig.migrate_package()
# or
mig3.migrate_package()

# End to end migration
mig.perform_full_migration()

# Detailed version
# Create import package and save it to the file specified with `save_path`
# argument during creation of migration object
mig.create_package()
mig2.create_package()

# Migrate downloaded package to the target environment
# `migrate_package()` by default uses a package binary saved to a variable
# during `create_package()`
mig.migrate_package()

# or a custom package binary specified with `custom_package_path`
# during Migration object creation, if `create_package()` was not called.
mig3.migrate_package()

# It is also possible to respecify `custom_package_path` at this stage
mig3.migrate_package(custom_package_path=CUSTOM_PACKAGE_PATH)

# Perform end to end migration. `perform_full_migration()` encapsulates
# `create_package()` and `migrate_package()` from the previous steps
# In order to be able to use, Migration object needs `source_connection`,
# `configuration` and `target_connection` parameters filled during creation
mig.perform_full_migration()

# Perform many full migrations at once
bulk_full_migration([mig])

# Perform many migrations at once
bulk_migrate_package([mig, mig3])

# If the migration needs to be reverted use `undo_migration()`
mig.undo_migration()

# Define a variable which can be later used in a script
UNDO_PACKAGE_PATH = $undo_package_path

# or run `migrate_package()` with path to the custom undo package
Migration(
    save_path=SAVE_PATH, target_connection=target_conn, custom_package_path=UNDO_PACKAGE_PATH
).migrate_package()

# Status of the migration can be checked by checking the `status` property
status = mig.status



"""This is the demo script to show how administrator can manage users and
user groups.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
import csv
from code_snippets.document import USER_ID

from mstrio.users_and_groups import (
    create_users_from_csv, list_user_groups, list_users, User, UserGroup
)

from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
USERNAME_1 = $username_1
FULLNAME_1 = $full_name_1
USERNAME_2 = $username_2
FULLNAME_2 = $full_name_2
USERNAME_3 = $username_3
FULLNAME_3 = $full_name_3
USERNAME_4 = $username_4
FULLNAME_4 = $full_name_4
USERNAME_5 = $username_5
FULLNAME_5 = $full_name_5

# Note: To create user with no password you must set minimum password length as
# 0 in environment security settings
# create multiple users
users_array = [
    {
        'username': USERNAME_1, 'fullName': FULLNAME_1
    }, {
        'username': USERNAME_2, 'fullName': FULLNAME_2
    }, {
        'username': USERNAME_3, 'fullName': FULLNAME_3
    }, {
        'username': USERNAME_4, 'fullName': FULLNAME_4
    }, {
        'username': USERNAME_5, 'fullName': FULLNAME_5
    }
]
for u in users_array:
    User.create(connection=conn, username=u['username'], full_name=u['fullName'])

# Define a variable which can be later used in a script
CSV_FILE = $csv_file  # csv file to create user from

# Also, you can create users from a CSV file
newly_created_users = create_users_from_csv(connection=conn, csv_file=CSV_FILE)

# Note: To create user with no password you must set minimum password length as
# 0 in environment security settings
# Or you can do it manually
with open(CSV_FILE, "r") as f:
    users = csv.DictReader(f)

    for user in users:
        User.create(connection=conn, username=user['username'], full_name=user['full_name'])

# Define variables which can be later used in a script
USERNAME_6 = $username_6
FULLNAME_6 = $full_name_6
NAME_BEGINS = $name_begins  # beginning of the full name to look for in listing users
INITIALS = $initials # initials to look for in listing users

# Note: To create user with no password you must set minimum password length as
# 0 in environment security settings
# create a single user without password and get users which name begins with
# "John" and have additional filter for initials
User.create(connection=conn, username=USERNAME_6, full_name=FULLNAME_6)
my_users = list_users(connection=conn, name_begins=NAME_BEGINS, initials=INITIALS)

# Define variables which can be later used in a script
USERNAME_7 = $username_7
FULLNAME_7 = $full_name_7
PASSWORD_7 = $password_7

# create a single user with password already defined
User.create(connection=conn, username=USERNAME_7, full_name=FULLNAME_7, password=PASSWORD_7)

# Define a variable which can be later used in a script
USER_GROUP_NAME = $user_group_name  # name of a User Group to create and add users
USER_GROUP_ID = $user_group_id
USER_ID = $user_id

# get all user groups (you can also add additional filters as for users) and
# create a new one
user_groups_list = list_user_groups(connection=conn)
UserGroup.create(connection=conn, name=USER_GROUP_NAME)

# get user by name
user = User(connection=conn, name=FULLNAME_6)
# get user by id
user = User(connection=conn, id=USER_ID)

# get user group by name
user_group = UserGroup(connection=conn, name=USER_GROUP_NAME)
# get user group by id
user_group = UserGroup(connection=conn, id=USER_GROUP_ID)
# add user to user_group
user_group.add_users(users=[user.id])

# Available values for permissions are 'grant', 'deny', 'default' or None
grant_right = 'right'

# Define variables which can be later used in a script
OBJECT_ID = $object_id  # ID of an object to which permissions will be set
OBJECT_TYPE = $object_type  # type of an object to which permissions will be set;
# see ObjectTypes enum in types.py for available object type values

# set custom permissions of the user for given objects
user.set_custom_permissions(
    to_objects=[OBJECT_ID],
    object_type=OBJECT_TYPE,
    execute=grant_right,
    use=grant_right,
    control=grant_right,
    delete=grant_right,
    write=grant_right,
    read=grant_right,
    browse=grant_right
)

# delete a user and a user group
user.delete()
user_group.delete()

# Addresses
ADDRESS_NAME = $address_name
ADDRESS = $address
user_john = User(connection=conn, name=FULLNAME_6)
# The user's addresses list is lazily loaded upon first accessing the property
johns_addresses = user_john.addresses
# MicroStrategy allows having multiple addresses marked as default
# as long as they are assigned to different device.
# The 'default' argument is automatically set to True.
# Let's add new default address to John.
user_john.add_address(name=ADDRESS_NAME, address=ADDRESS)
# It can also be achieved by setting value of 'default' argument to True.
user_john.add_address(name=ADDRESS_NAME, address=ADDRESS, default=True)
# If the address should not be default, the 'default' should be set to False.
user_john.add_address(name=ADDRESS_NAME, address=ADDRESS, default=False)
# Let's construct a list of John's "default" addresses
johns_default_addresses = [addr for addr in johns_addresses if addr["is_default"]]
# If we just want a single address to treat as default and we're OK with
# it being just the first "default" one we can now just take the first element
# of the list
johns_default_address = johns_default_addresses[0]
# Alternatively if we don't care about any addresses besides the first
# "default one" we can use a generator expression and take its first output
# to avoid going through the entire list, increasing performance
johns_default_address = next((addr for addr in user_john.addresses if addr["is_default"]), None)
# By default, when only ADDRESS_NAME and ADDRESS are provided, an email type
# address is created. It is possible to create more address types by providing
# DEVICE_ID and DELIVERY_TYPE
ADDRESS_NAME = $address_name
ADDRESS = $address
DEVICE_ID = $device_id
DELIVERY_TYPE = $delivery_type
user_john.add_address(
    name=ADDRESS_NAME,
    address=ADDRESS,
    device_id=DEVICE_ID,
    delivery_type=DELIVERY_TYPE
)
# It is possible to make changes to an existing address
ADDRESS_ID = $address_id
NEW_ADDRESS_NAME = $new_address_name
NEW_ADDRESS = $new_address
NEW_DEVICE_ID = $new_device_id
NEW_DELIVERY_TYPE = $new_delivery_type
user_john.update_address(
    id=ADDRESS_ID,
    name=NEW_ADDRESS_NAME,
    address=NEW_ADDRESS,
    device_id=NEW_DEVICE_ID,
    delivery_type=NEW_DELIVERY_TYPE
)
